from typing import Optional
import torch


class StaticTensor:
    """
    A tensor holder that is allocated only once

    This is useful for reducing the memory allocator overhead, and is required
    for input and output tensors of a CUDA graph.

    Initialize a `StaticTensor` with an initial tensor and/or the maximum buffer
    size (in number of elements). Then, use `.set()` and `.get()` to set and get
    the tensor. The tensor will NOT be reallocated during `.set()`.

    Args:
        tensor: The initial tensor to be stored. Defaults to an empty tensor.
        max_nelem: The maximum number of elements that the tensor can store.
            Defaults to the number of elements in the initial tensor.
        dtype: The data type of the tensor. Defaults to the data type of the
            initial tensor.
        device: The device of the tensor. Defaults to the device of the initial
            tensor.
    """

    def __init__(
        self,
        tensor: Optional[torch.Tensor] = None,
        *,
        max_nelem: Optional[int] = None,
        dtype: Optional[torch.device] = None,
        device: Optional[torch.device] = None,
    ):
        if max_nelem is None:
            if tensor is None:
                raise ValueError(f"max_nelem must be specified if tensor is None")
            max_nelem = tensor.numel()
        if dtype is None:
            if tensor is None:
                raise ValueError(f"dtype must be specified if tensor is None")
            dtype = tensor.dtype
        if device is None:
            if tensor is None:
                raise ValueError(f"device must be specified if tensor is None")
            device = tensor.device

        self._buffer = torch.empty(max_nelem, dtype=dtype, device=device)
        self._cur_nelem = 0
        self._cur_shape = (0,)

        if tensor is not None:
            self.set(tensor)

    def set(self, tensor: torch.Tensor):
        """
        Set the tensor to be stored

        Args:
            tensor: The tensor to be stored
        """
        if tensor.numel() > self._buffer.numel():
            raise ValueError(
                f"The assigned tensor ({tensor.shape}) cannot be larger than the buffer ({self._buffer.shape})"
            )
        if tensor.dtype != self._buffer.dtype:
            raise ValueError(
                f"The assigned tensor ({tensor.dtype}) must have the same data type as the buffer ({self._buffer.dtype})"
            )
        if tensor.device != self._buffer.device:
            raise ValueError(
                f"The assigned tensor ({tensor.device}) must have the same device as the buffer ({self._buffer.device})"
            )
        self._buffer[: tensor.numel()].copy_(tensor.flatten())
        self._cur_nelem = tensor.numel()
        self._cur_shape = tensor.shape

    def get(self) -> torch.Tensor:
        """
        Get the stored tensor

        Returns:
            The stored tensor
        """
        return self._buffer[: self._cur_nelem].reshape(self._cur_shape)



import time
from logging import getLogger
from typing import List  # Please keep Python 3.8 compatible

import torch

from chitu.global_vars import get_global_args, get_slot_handle
from chitu.task import TaskPool, TaskType

logger = getLogger(__name__)


class Scheduler:
    @staticmethod
    def build(args):
        if args.type.lower() == "fifo" or args.type.lower() == "fcfs":
            return FcfsScheduler(args.fcfs.num_tasks, args.fcfs.enable_hybrid)
        if args.type.lower() == "prefill_first":
            slot_handle = get_slot_handle()
            if slot_handle:
                return SkewPipelineScheduler(
                    args.prefill_first.num_tasks, args.prefill_first.enable_hybrid
                )
            else:
                return PrefillFirstScheduler(
                    args.prefill_first.num_tasks, args.prefill_first.enable_hybrid
                )
        if args.type.lower() == "stride":
            return StrideScheduler(args.stride.num_tasks, args.stride.enable_hybrid)
        if args.type.lower() == "deadline":
            return DdlScheduler(args.deadline.num_tasks, args.deadline.enable_hybrid)
        if args.type.lower() == "prefix_align":
            return PrefixAlignScheduler(
                args.prefix_align.num_tasks, args.prefix_align.enable_hybrid
            )
        if args.type.lower() == "balance":
            return BalanceScheduler(args.balance.num_tasks, args.balance.enable_hybrid)
        else:
            raise NotImplementedError(f"Scheduler {args.type} not implemented")

    def schedule(self) -> List[str]:
        raise NotImplementedError()

    def update(self, cur_task_ids, unwait_task_ids=[]):
        removed_task_ids = []
        task_ids = cur_task_ids + unwait_task_ids
        task_ids = list(set(task_ids))
        for task_id in task_ids:
            if TaskPool.pool[task_id].need_remove():
                if TaskPool.pool[task_id].task_type == TaskType.Decode:
                    removed_task_ids.append(task_id)
                assert TaskPool.remove(
                    task_id
                ), f"Task {task_id} not found in pool {TaskPool.pool.keys()}"
            # assert False
        # for task_id in unwait_task_ids:
        #     if task_id in TaskPool.id_list and TaskPool.pool[task_id].need_remove():
        #         assert TaskPool.remove(task_id), "Task not found in pool"
        return removed_task_ids

    def is_done(self):
        return len(TaskPool.pool) == 0


class FcfsScheduler(Scheduler):
    """
    first come, first service
    note that no arrival_time record, implicitly ordered by TaskPool.add -> list.append
    """

    def __init__(self, num_tasks: int, enable_hybrid: bool):
        super().__init__()
        assert num_tasks > 0, "num_tasks must be greater than 0"
        self.num_tasks = num_tasks
        self.enable_hybrid = enable_hybrid

    def schedule(self) -> List[str]:
        if self.enable_hybrid:
            ret_task_ids = TaskPool.id_list[: self.num_tasks]
        else:
            filter_task_type = (
                TaskPool.pool[TaskPool.id_list[0]].task_type
                if len(TaskPool.id_list) > 0
                else TaskType.Prefill
            )
            filtered_task_ids = filter(
                lambda x: TaskPool.pool[x].task_type == filter_task_type,
                TaskPool.id_list,
            )
            ret_task_ids = list(filtered_task_ids)[: self.num_tasks]
        if filter_task_type == TaskType.Prefill:
            ret_task_ids = ret_task_ids[:1]
        logger.debug(f"Selected task_ids: {self.ret_task_ids}")
        return ret_task_ids


class PrefillFirstScheduler(Scheduler):
    """
    always select prefill tasks, in a fifo manner, if any.
    decode tasks will be selected only if no prefill task
    """

    def __init__(self, num_tasks: int, enable_hybrid: bool):
        super().__init__()
        assert num_tasks > 0, "num_tasks must be greater than 0"
        self.num_tasks = num_tasks
        self.enable_hybrid = enable_hybrid

    def schedule(self) -> List[str]:
        prefill_task_ids = filter(
            lambda x: TaskPool.pool[x].task_type == TaskType.Prefill
            and not TaskPool.pool[x].waiting,
            TaskPool.id_list,
        )
        # select at most num_tasks prefill tasks
        ret_task_ids = list(prefill_task_ids)[: self.num_tasks]
        # if no prefill tasks or enable hybrid, select decode tasks if there is room left
        if len(ret_task_ids) == 0 or (
            self.enable_hybrid and len(ret_task_ids) < self.num_tasks
        ):
            decode_task_ids = filter(
                lambda x: TaskPool.pool[x].task_type == TaskType.Decode
                and not TaskPool.pool[x].waiting,
                TaskPool.id_list,
            )
            ret_task_ids.extend(
                list(decode_task_ids)[: self.num_tasks - len(ret_task_ids)]
            )
        logger.debug(f"Selected task_ids: {ret_task_ids}")

        # if (
        #     len(ret_task_ids) > 0
        #     and TaskPool.pool[ret_task_ids[0]].task_type == TaskType.Prefill
        # ):
        #     ret_task_ids = ret_task_ids[:1]
        return ret_task_ids


class SkewPipelineScheduler(Scheduler):

    def __init__(self, num_tasks: int, enable_hybrid: bool):
        super().__init__()
        assert num_tasks > 0, "num_tasks must be greater than 0"
        self.num_tasks = num_tasks
        self.enable_hybrid = enable_hybrid
        self.slot_handle = get_slot_handle()
        self.decode_slots = [[] for _ in range(self.slot_handle.num_slots)]
        self.slot_id = 0

    def schedule(self) -> List[str]:
        #  search for an empty slot to prefill
        prefill_task_ids = filter(
            lambda x: TaskPool.pool[x].task_type == TaskType.Prefill
            and not TaskPool.pool[x].waiting,
            TaskPool.id_list,
        )

        local_idx = -1
        for idx, slots in enumerate(self.decode_slots):
            if len(slots) == 0:
                local_idx = idx
                break

        num_tasks = 0 if local_idx == -1 else self.slot_handle.get_slot_size(local_idx)
        ret_task_ids = list(prefill_task_ids)[:num_tasks]

        if num_tasks:
            self.decode_slots[local_idx].extend(ret_task_ids)

        decode_task_ids = []
        if len(ret_task_ids) == 0 or (
            self.enable_hybrid and len(ret_task_ids) < self.num_tasks
        ):
            for idx, slot_ids in enumerate(self.decode_slots):
                if len(slot_ids) > 0 and not TaskPool.pool[slot_ids[0]].waiting:
                    local_idx = idx
                    decode_task_ids = slot_ids
                    break
            ret_task_ids.extend(
                list(decode_task_ids)[: self.num_tasks - len(ret_task_ids)]
            )

        if len(ret_task_ids):
            self.slot_handle.set_slot_idx(local_idx)

        return ret_task_ids


class StrideScheduler(Scheduler):
    """
    each task has a priority value P, and a score S (starts from 0),
    at scheduling point, update the scores:
    S += P * elapsed_time
    select the tasks with top scores and reset their scores back to 0.
    """

    def __init__(self, num_tasks: int, enable_hybrid: bool):
        super().__init__()
        assert num_tasks > 0, "num_tasks must be greater than 0"
        self.num_tasks = num_tasks
        self.enable_hybrid = enable_hybrid

    def schedule(self) -> List[str]:
        # update sched_score
        for task_id in TaskPool.id_list:
            task = TaskPool.pool[task_id]
            task.sched_score += task.priority * (time.perf_counter_ns() - task.sched_ts)
            task.sched_ts = time.perf_counter_ns()
        # sort by sched_score and select top tasks
        if self.enable_hybrid:
            ret_task_ids = sorted(
                TaskPool.id_list,
                key=lambda x: TaskPool.pool[x].sched_score,
                reverse=True,
            )[: self.num_tasks]
        else:
            filter_task_type = (
                TaskPool.pool[TaskPool.id_list[0]].task_type
                if len(TaskPool.id_list) > 0
                else TaskType.Prefill
            )
            filtered_task_ids = filter(
                lambda x: TaskPool.pool[x].task_type == filter_task_type,
                TaskPool.id_list,
            )
            ret_task_ids = sorted(
                list(filtered_task_ids),
                key=lambda x: TaskPool.pool[x].sched_score,
                reverse=True,
            )[: self.num_tasks]
        # reset sched_score of selected tasks
        for task_id in ret_task_ids:
            TaskPool.pool[task_id].sched_score = 0
        logger.debug(f"Selected task_ids: {ret_task_ids}")
        return ret_task_ids


class DdlScheduler(Scheduler):
    """
    each task has a deadline time DDL
    DDL = request_arrival_time + prefix_length*alpha + max_output_tokens*beta
    select the tasks with nearest DDL.
    alpha and beta are arbitary value, defaults to 1ms.
    """

    def __init__(self, num_tasks: int, enable_hybrid: bool):
        super().__init__()
        assert num_tasks > 0, "num_tasks must be greater than 0"
        self.num_tasks = num_tasks
        self.enable_hybrid = enable_hybrid

    def schedule(self) -> List[str]:
        # sort by ddl and select top tasks
        if self.enable_hybrid:
            ret_task_ids = sorted(
                TaskPool.id_list, key=lambda x: TaskPool.pool[x].sched_ddl
            )[: self.num_tasks]
        else:
            filter_task_type = (
                TaskPool.pool[TaskPool.id_list[0]].task_type
                if len(TaskPool.id_list) > 0
                else TaskType.Prefill
            )
            filtered_task_ids = filter(
                lambda x: TaskPool.pool[x].task_type == filter_task_type,
                TaskPool.id_list,
            )
            ret_task_ids = sorted(
                list(filtered_task_ids), key=lambda x: TaskPool.pool[x].sched_ddl
            )[: self.num_tasks]
        logger.debug(f"Selected task_ids: {ret_task_ids}")
        return ret_task_ids


class PrefixAlignScheduler(Scheduler):
    """
    each task has a prefix length, for a prefill task, prefix = tokenized prompt,
    for a decode task, prefix = prefill's prefix + generated tokens.
    at scheduling point, try to select as many tasks with close prefix_length as possible.
    E.g., 4 tasks with prefix_length 100 and 2 tasks with prefix_length 300, select the former 4 tasks.
    """

    def __init__(self, num_tasks: int, enable_hybrid: bool):
        super().__init__()
        assert num_tasks > 0, "num_tasks must be greater than 0"
        self.num_tasks = num_tasks
        self.enable_hybrid = enable_hybrid

    def schedule(self) -> List[str]:
        # TODO: no definition on 'close prefix length', sort by prefix length and select the longest
        if self.enable_hybrid:
            ret_task_ids = sorted(
                TaskPool.id_list, key=lambda x: TaskPool.pool[x].prefix_length
            )[: self.num_tasks]
        else:
            filter_task_type = (
                TaskPool.pool[TaskPool.id_list[0]].task_type
                if len(TaskPool.id_list) > 0
                else TaskType.Prefill
            )
            filtered_task_ids = filter(
                lambda x: TaskPool.pool[x].task_type == filter_task_type,
                TaskPool.id_list,
            )
            ret_task_ids = sorted(
                list(filtered_task_ids), key=lambda x: TaskPool.pool[x].prefix_length
            )[: self.num_tasks]
        logger.debug(f"Selected task_ids: {ret_task_ids}")
        return ret_task_ids


class BalanceScheduler(Scheduler):
    """
    an advanced scheduler that pack appropriate prefill tasks and decode tasks to
    acheive the best hardware utilization. since prefill tasks are computation-intensive
    while decode tasks are memory-intensive, there must be a best hybrid solution.
    however, searching for the optimal solution is expensive,
    so a somehow heuristic algorithm will be applied.
    """

    def __init__(self, num_tasks: int, enable_hybrid: bool):
        super().__init__()
        assert num_tasks > 0, "num_tasks must be greater than 0"
        self.num_tasks = num_tasks
        self.enable_hybrid = enable_hybrid

    def schedule(self) -> List[str]:
        prefill_task_ids = list(
            filter(
                lambda x: TaskPool.pool[x].task_type == TaskType.Prefill,
                TaskPool.id_list,
            )
        )
        decode_task_ids = list(
            filter(
                lambda x: TaskPool.pool[x].task_type == TaskType.Decode,
                TaskPool.id_list,
            )
        )
        if self.enable_hybrid:  # TODO: currently half and half
            prefill_count = (
                self.num_tasks // 2
                if len(decode_task_ids) >= self.num_tasks // 2
                else self.num_tasks - len(decode_task_ids)
            )
            decode_count = self.num_tasks - prefill_count
            ret_task_ids = prefill_task_ids[:prefill_count]
            ret_task_ids.extend(decode_task_ids[:decode_count])
        else:  # fall back to prefill_first
            ret_task_ids = (
                prefill_task_ids[: self.num_tasks]
                if len(prefill_task_ids) > 0
                else decode_task_ids[: self.num_tasks]
            )
        logger.debug(f"Selected task_ids: {ret_task_ids}")
        return ret_task_ids



from logging import getLogger

import torch

from chitu.models.model import *
from chitu.tensor_parallel import ColumnParallelLinear, RowParallelLinear
from chitu.tokenizer import ChatFormat, Tokenizer

logger = getLogger(__name__)


__all__ = ["quant"]


def replace_with_bnb(model, current_key_name=None):

    import bitsandbytes as bnb

    has_been_replaced = False
    for name, module in model.named_children():
        if current_key_name is None:
            current_key_name = []

        current_key_name.append(name)
        current_key_name_str = ".".join(current_key_name)
        # if current_key_name_str == "model.model":
        #    continue
        # if (name == "model"):
        #    continue
        if isinstance(
            module, (torch.nn.Linear, ColumnParallelLinear, RowParallelLinear)
        ):
            bnb_module = bnb.nn.Linear8bitLt(
                module.in_features,
                module.out_features,
                module.bias is not None,
                has_fp16_weights=False,
                threshold=6.0,
            )
            bnb_module.weight.data = module.weight.data
            if module.bias is not None:
                bnb_module.bias.data = module.bias.data
            bnb_module.requires_grad_(False)
            setattr(model, name, bnb_module)
            has_been_replaced = True
        if len(list(module.children())) > 0:
            _, _has_been_replaced = replace_with_bnb(module, current_key_name)
            has_been_replaced = has_been_replaced | _has_been_replaced

        current_key_name.pop(-1)
    return model, has_been_replaced


def replace_with_gptq(model, current_key_name=None):

    has_been_replaced = False
    for name, module in model.named_children():
        if current_key_name is None:
            current_key_name = []

        current_key_name.append(name)
        current_key_name_str = ".".join(current_key_name)
        if name != "lm_head":
            if isinstance(
                module, (torch.nn.Linear, ColumnParallelLinear, RowParallelLinear)
            ):
                from auto_gptq.nn_modules.qlinear.qlinear_tritonv2 import QuantLinear

                gptq_linear = QuantLinear(
                    8,
                    128,
                    module.in_features,
                    module.out_features,
                    module.bias is not None,
                    trainable=False,
                    weight_dtype=module.weight.dtype,
                )
                gptq_linear.requires_grad_(False)
                setattr(model, name, gptq_linear)
                has_been_replaced = True
            if len(list(module.children())) > 0:
                _, _has_been_replaced = replace_with_gptq(module, current_key_name)
                has_been_replaced = has_been_replaced | _has_been_replaced

        current_key_name.pop(-1)
    return model, has_been_replaced


def replace_with_w8a16(model, current_key_name=None):

    has_been_replaced = False
    for name, module in model.named_children():
        if current_key_name is None:
            current_key_name = []

        current_key_name.append(name)
        current_key_name_str = ".".join(current_key_name)
        if name != "lm_head":
            if isinstance(
                module, (torch.nn.Linear, ColumnParallelLinear, RowParallelLinear)
            ):
                from chitu.quantize.w8a16 import WeightOnlyLinear

                w8a16_linear = WeightOnlyLinear(
                    module.in_features, module.out_features, module.bias is not None
                )
                w8a16_linear.requires_grad_(False)
                setattr(model, name, w8a16_linear)
                has_been_replaced = True
            if len(list(module.children())) > 0:
                _, _has_been_replaced = replace_with_w8a16(module, current_key_name)
                has_been_replaced = has_been_replaced | _has_been_replaced

        current_key_name.pop(-1)
    return model, has_been_replaced


def replace_with_simple_w8a8(model, current_key_name=None, quant_on_load=False):

    has_been_replaced = False
    for name, module in model.named_children():
        if current_key_name is None:
            current_key_name = []

        current_key_name.append(name)
        current_key_name_str = ".".join(current_key_name)
        if name != "lm_head":
            if isinstance(
                module, (torch.nn.Linear, ColumnParallelLinear, RowParallelLinear)
            ):
                from chitu.quantize.w8a8 import W8A8Linear

                w8a8_linear = W8A8Linear.from_float(
                    module, model_arch_only=not quant_on_load
                )
                w8a8_linear.requires_grad_(False)
                setattr(model, name, w8a8_linear)
                has_been_replaced = True
            if len(list(module.children())) > 0:
                _, _has_been_replaced = replace_with_simple_w8a8(
                    module, current_key_name, quant_on_load=quant_on_load
                )
                has_been_replaced = has_been_replaced | _has_been_replaced

        current_key_name.pop(-1)
    return model, has_been_replaced


def replace_with_simple_w8a8_muxi(model, current_key_name=None, quant_on_load=False):

    has_been_replaced = False
    for name, module in model.named_children():
        if current_key_name is None:
            current_key_name = []

        current_key_name.append(name)
        current_key_name_str = ".".join(current_key_name)
        if name != "lm_head":
            from chitu.models.model import RMSNorm
            from chitu.quantize.muxi_w8a8 import NormAndQuant, W8A8Linear

            if isinstance(
                module, (torch.nn.Linear, ColumnParallelLinear, RowParallelLinear)
            ):

                w8a8_linear = W8A8Linear.from_float(
                    module, model_arch_only=not quant_on_load
                )
                w8a8_linear.requires_grad_(False)
                setattr(model, name, w8a8_linear)
                has_been_replaced = True

            if isinstance(module, (RMSNorm)) and name != "norm":
                new_norm = NormAndQuant.from_float(
                    module, model_arch_only=not quant_on_load
                )
                new_norm.requires_grad_(False)
                setattr(model, name, new_norm)
                has_been_replaced = True

            if len(list(module.children())) > 0:
                _, _has_been_replaced = replace_with_simple_w8a8_muxi(
                    module, current_key_name, quant_on_load=quant_on_load
                )
                has_been_replaced = has_been_replaced | _has_been_replaced

        current_key_name.pop(-1)
    return model, has_been_replaced


def quantize_llmint8(model):
    model, has_been_replaced = replace_with_bnb(model)

    if not has_been_replaced:
        logger.warning("No linear modules were quantized.")

    logger.debug(f"Quantized model: {model}")
    return model


def quantize_gptq(model):
    model, has_been_replaced = replace_with_gptq(model)

    if not has_been_replaced:
        logger.warning("No linear modules were quantized.")

    logger.debug(f"Quantized model: {model}")
    return model


def quantize_awq(model, name="hf-llama"):
    q_config = {"zero_point": True, "q_group_size": 128}

    from chitu import awq

    awq.real_quantize_model_weight(model, w_bit=4, q_config=q_config, init_only=True)

    if name == "llama":
        sd = torch.load(
            "/home/tanyijun/cinfer/quant_cache/Llama3-8B-4bit.pth", map_location="cpu"
        )
        model.load_state_dict(sd)
    elif name == "hf-llama":
        0

    class FP16Trans(torch.nn.Module):
        def __init__(self, toke):
            super().__init__()
            self.tok_embeddings = toke

        def forward(self, x):
            return self.tok_embeddings(x).to(torch.float16)

    if name == "llama":
        model.tok_embeddings = FP16Trans(model.tok_embeddings)
    elif name == "hf-llama":
        model.embed_tokens = FP16Trans(model.embed_tokens)

    logger.debug(f"Quantized model: {model}")

    return model


def quantize_w8a16(model):
    model, has_been_replaced = replace_with_w8a16(model)

    if not has_been_replaced:
        raise NotImplementedError("error load model")

    logger.debug(f"Quantized model: {model}")
    return model


def quantize_simple_w8a8(model, quant_on_load=False):
    model, has_been_replaced = replace_with_simple_w8a8(
        model, quant_on_load=quant_on_load
    )

    if not has_been_replaced:
        raise NotImplementedError("error load model")

    logger.debug(f"Quantized model: {model}")
    return model


def quantize_simple_w8a8_muxi(model, quant_on_load=False):
    model, has_been_replaced = replace_with_simple_w8a8_muxi(
        model, quant_on_load=quant_on_load
    )

    if not has_been_replaced:
        raise NotImplementedError("error load model")

    logger.debug(f"Quantized model: {model}")
    return model


def quant(model, method=None, name="hf-llama", quant_on_load=False):

    if method == "llmint8":
        return quantize_llmint8(model)
    elif method == "awq":
        return quantize_awq(model, name)
    elif method == "gptq":
        return quantize_gptq(model)
    elif method == "w8a16":
        return quantize_w8a16(model)
    elif method == "simple_w8a8":
        return quantize_simple_w8a8(model, quant_on_load=quant_on_load)
    elif method == "simple_w8a8_muxi":
        return quantize_simple_w8a8_muxi(model, quant_on_load=quant_on_load)
    return model



import os
import sys
import time
from functools import reduce
import operator
import torch
from logging import getLogger

logger = getLogger(__name__)


_GLOBAL_ARGS = None
_GLOBAL_NUM_MICROBATCHES_CALCULATOR = None
_GLOBAL_TOKENIZER = None
_GLOBAL_TENSORBOARD_WRITER = None
_GLOBAL_ADLR_AUTORESUME = None
_GLOBAL_TIMERS = None
_GLOBAL_SIGNAL_HANDLER = None
_GLOBAL_MEMORY_BUFFER = None
_GLOBAL_SLOT_HANDLE = None


def get_global_memory_buffer():
    _ensure_var_is_initialized(_GLOBAL_MEMORY_BUFFER, "global memory buffer")
    return _GLOBAL_MEMORY_BUFFER


def get_slot_handle():
    # _ensure_var_is_initialized(_GLOBAL_SLOT_HANDLE, "slot_handle")
    return _GLOBAL_SLOT_HANDLE


def set_global_variables(global_args=None):
    set_global_args(global_args)
    _set_timers()
    if global_args is not None:
        _set_slot_handle(
            global_args.infer.max_reqs,
            global_args.infer.pp_size,
            global_args.infer.cache_type,
        )


def _set_slot_handle(max_reqs, pp_size, cache_type):
    global _GLOBAL_SLOT_HANDLE
    _ensure_var_is_not_initialized(_GLOBAL_SLOT_HANDLE, "slot_handle")
    if cache_type == "skew" and pp_size > 1:
        _GLOBAL_SLOT_HANDLE = SlotHandle(max_reqs, pp_size)


def _set_tensorboard_writer(args):
    """Set tensorboard writer."""
    global _GLOBAL_TENSORBOARD_WRITER
    _ensure_var_is_not_initialized(_GLOBAL_TENSORBOARD_WRITER, "tensorboard writer")

    if (
        hasattr(args, "tensorboard_dir")
        and args.tensorboard_dir
        and args.rank == (args.world_size - 1)
    ):
        try:
            from torch.utils.tensorboard import SummaryWriter

            logger.info("> setting tensorboard ...")
            _GLOBAL_TENSORBOARD_WRITER = SummaryWriter(
                log_dir=args.tensorboard_dir, max_queue=args.tensorboard_queue_size
            )
        except ModuleNotFoundError:
            logger.warning(
                "TensorBoard writing requested but is not "
                "available (are you using PyTorch 1.1.0 or later?), "
                "no TensorBoard logs will be written.",
                flush=True,
            )


def get_global_args():
    _ensure_var_is_initialized(_GLOBAL_ARGS, "global args")
    return _GLOBAL_ARGS


def set_global_args(args):
    global _GLOBAL_ARGS
    _ensure_var_is_not_initialized(_GLOBAL_ARGS, "global args")
    _GLOBAL_ARGS = args


def get_timers():
    """Return timers."""
    _ensure_var_is_initialized(_GLOBAL_TIMERS, "timers")
    return _GLOBAL_TIMERS


def _set_timers():
    """Initialize timers."""
    global _GLOBAL_TIMERS
    _ensure_var_is_not_initialized(_GLOBAL_TIMERS, "timers")
    _GLOBAL_TIMERS = Timers()


def _set_global_memory_buffer():
    """Initialize global buffer"""
    global _GLOBAL_MEMORY_BUFFER
    _ensure_var_is_not_initialized(_GLOBAL_MEMORY_BUFFER, "global memory buffer")
    _GLOBAL_MEMORY_BUFFER = GlobalMemoryBuffer()


def _ensure_var_is_initialized(var, name):
    """Make sure the input variable is not None."""
    assert var is not None, "{} is not initialized.".format(name)


def _ensure_var_is_not_initialized(var, name):
    """Make sure the input variable is not None."""
    assert var is None, "{} is already initialized.".format(name)


class _Timer:
    """Timer."""

    def __init__(self, name):
        self.name_ = name
        self.elapsed_ = 0.0
        self.started_ = False
        self.start_time = time.time()
        self.cnt = 0

    def start(self):
        # return
        """Start the timer."""
        assert not self.started_, "timer has already been started"
        torch.cuda.synchronize()
        self.start_time = time.time()
        self.started_ = True

    def stop(self):
        # return
        """Stop the timer."""
        assert self.started_, "timer is not started"
        torch.cuda.synchronize()
        self.elapsed_ += time.time() - self.start_time
        self.started_ = False
        self.cnt += 1

    def reset(self):
        """Reset timer."""
        self.elapsed_ = 0.0
        self.started_ = False
        self.cnt = 0

    def elapsed(self, reset=True):
        """Calculate the elapsed time."""
        started_ = self.started_
        # If the timing in progress, end it first.
        if self.started_:
            self.stop()
        # Get the elapsed time.
        elapsed_ = self.elapsed_
        # Reset the elapsed time
        if reset:
            self.reset()
        # If timing was in progress, set it back.
        if started_:
            self.start()
        return elapsed_


class Timers:
    """Group of timers."""

    def __init__(self):
        self.timers = {}

    def __call__(self, name):
        if name not in self.timers:
            self.timers[name] = _Timer(name)
        return self.timers[name]

    def write(self, names, writer, iteration, normalizer=1.0, reset=False):
        """Write timers to a tensorboard writer"""
        # currently when using add_scalars,
        # torch.utils.add_scalars makes each timer its own run, which
        # polutes the runs list, so we just add each as a scalar
        assert normalizer > 0.0
        for name in names:
            value = self.timers[name].elapsed(reset=reset) / normalizer
            writer.add_scalar(name + "-time", value, iteration)

    def log(self, names=[], normalizer=1.0, reset=True):
        """Log a group of timers."""
        if len(names) == 0:
            names = self.timers.keys()
        assert normalizer > 0.0
        string = "time (ms)"
        for name in names:
            cnt = self.timers[name].cnt
            if cnt == 0:
                continue
            elapsed_time = self.timers[name].elapsed(reset=reset) * 1000.0 / normalizer
            string += " | {}: {:.2f} {} {:.2f}".format(
                name, elapsed_time, cnt, elapsed_time / cnt
            )
        logger.info(string)


class GlobalMemoryBuffer:
    """Global buffer to avoid dynamic memory allocations.
    Caller should ensure that buffers of the same name
    are not used concurrently."""

    def __init__(self):
        self.buffer = {}

    def get_tensor(self, tensor_shape, dtype, name):
        required_len = reduce(operator.mul, tensor_shape, 1)
        if (
            self.buffer.get((name, dtype), None) is None
            or self.buffer[(name, dtype)].numel() < required_len
        ):
            self.buffer[(name, dtype)] = torch.empty(
                required_len,
                dtype=dtype,
                device=torch.cuda.current_device(),
                requires_grad=False,
            )

        return self.buffer[(name, dtype)][0:required_len].view(*tensor_shape)


class SlotHandle:
    """
    split max_reqs to micro_batch size
    max_req = 10, pp_size = 3, self.slots_size = [4, 3, 3]
    """

    def __init__(self, max_reqs, pp_size):
        self.slots_size = self.split_slots(max_reqs, pp_size)
        self.num_slots = len(self.slots_size)
        self.slot_idx = 0
        self.slot_start_idx = []
        self.slot_end_idx = []

        res = [0]
        for value in self.slots_size:
            res.append(res[-1] + value)
        self.slot_start_idx = res[:-1]
        self.slot_end_idx = res[1:]

    def split_slots(self, total, parts):
        result = [0] * min(total, parts)
        for i in range(total):
            result[i % parts] += 1
        return result

    def get_slot_size(self, idx):
        return self.slots_size[idx]

    def set_slot_idx(self, idx):
        self.slot_idx = idx

    def get_slot_idx(self):
        return self.slot_idx

    def get_slot_start_end_idx(self, idx):
        return self.slot_start_idx[idx], self.slot_end_idx[idx]

    def get_current_slot_start_end_idx(self):
        return self.get_slot_start_end_idx(self.slot_idx)



from typing import Callable, Sequence, Mapping, Any, Optional
import functools
import torch

from chitu.static_tensor import StaticTensor


def make_dispatched_graphed_callables(
    f: Optional[Callable] = None,
    *,
    sample_args: Sequence[torch.Tensor],
    sample_kwargs: Mapping[str, torch.Tensor],
    args_max_nelem: Sequence[int],
    kwargs_max_nelem: Mapping[str, int],
    output_max_nelem_callback: Callable[[int], int],
    enable: bool = True,
) -> Callable:
    """
    Make a callable to run with CUDA graph but capature different graphs when `key` changes.

    Args:
        f: The function to wrap. Currently all the inputs should be tensors, and there should only be one
            output which is a tensor. If None, return a partial function as an decorator.
        sample_args: The sample positional arguments for capturing the graph.
        sample_kwargs: The sample keyword arguments for capturing the graph.
        args_max_nelem: The maximum number of elements in the positional arguments, used to hold inputs
            in shared static tensors.
        kwargs_max_nelem: The maximum number of elements in the keyword arguments, used to hold inputs
            in shared static tensors.
        output_max_nelem: A `(sample_nelem) -> max_nelem` callback to return the maximum number of elements
            in the output tensor, used to hold outputs in shared static tensors.
        enable: If False, do nothing but only add the `key` argument.

    Returns:
        The wrapped function, which has an additional first argument `key` to dispatch different graphs.
    """

    if f is None:
        return functools.partial(
            make_dispatched_graphed_callables,
            sample_args=sample_args,
            sample_kwargs=sample_kwargs,
            args_max_nelem=args_max_nelem,
            kwargs_max_nelem=kwargs_max_nelem,
            output_max_nelem_callback=output_max_nelem_callback,
            enable=enable,
        )

    if enable:

        graph_dict: Dict[Any, torch.cuda.CUDAGraph] = {}
        cuda_graph_pool = None

        args_static_tensors: Optinoal[List[StaticTensor]] = None
        kwargs_static_tensors: Optional[Dict[str, StaticTensor]] = None
        output_static_tensor: Optional[StaticTensor] = None

        def new_callable(key: Any, *args, **kwargs):
            nonlocal graph_dict
            nonlocal cuda_graph_pool
            nonlocal args_static_tensors
            nonlocal kwargs_static_tensors
            nonlocal output_static_tensor

            if key not in graph_dict:
                # Warmup
                sample_output = f(*sample_args, **sample_kwargs)

                # Allocate static tensors
                if args_static_tensors is None:
                    args_static_tensors = [
                        StaticTensor(sample, max_nelem=max_nelem)
                        for sample, max_nelem in zip(sample_args, args_max_nelem)
                    ]
                if kwargs_static_tensors is None:
                    kwargs_static_tensors = {}
                    for k in sample_kwargs:
                        kwargs_static_tensors[k] = StaticTensor(
                            sample_kwargs[k], max_nelem=kwargs_max_nelem[k]
                        )
                if output_static_tensor is None:
                    output_static_tensor = StaticTensor(
                        sample_output,
                        max_nelem=output_max_nelem_callback(sample_output.numel()),
                    )

                # Capture the graph
                graph_dict[key] = torch.cuda.CUDAGraph()
                with torch.cuda.graph(graph_dict[key], pool=cuda_graph_pool):
                    output = f(
                        *[static_tensor.get() for static_tensor in args_static_tensors],
                        **{
                            k: static_tensor.get()
                            for k, static_tensor in kwargs_static_tensors.items()
                        },
                    )
                    output_static_tensor.set(output)
                if cuda_graph_pool is None:
                    cuda_graph_pool = graph_dict[key].pool()

            graph_dict[key].replay()
            return output_static_tensor.get()

    else:  # not enable

        def new_callable(key: Any, *args, **kwargs):
            return f(*args, **kwargs)

    return functools.update_wrapper(new_callable, f)



import torch
import torch.nn as nn

import gc

__all__ = ["auto_clip_block", "apply_clip"]


def pseudo_quantize_tensor(
    w, n_bit=8, zero_point=True, q_group_size=-1, inplace=False, get_scale_zp=False
):
    org_w_shape = w.shape
    if q_group_size > 0:
        assert org_w_shape[-1] % q_group_size == 0
        w = w.reshape(-1, q_group_size)
    assert w.dim() == 2
    if zero_point:
        max_val = w.amax(dim=1, keepdim=True)
        min_val = w.amin(dim=1, keepdim=True)
        max_int = 2**n_bit - 1
        min_int = 0
        scales = (max_val - min_val).clamp(min=1e-5) / max_int
        zeros = (-torch.round(min_val / scales)).clamp_(min_int, max_int)

    assert torch.isnan(scales).sum() == 0
    assert torch.isnan(w).sum() == 0

    if inplace:
        (
            (w.div_(scales).round_().add_(zeros)).clamp_(min_int, max_int).sub_(zeros)
        ).mul_(scales)
    else:
        w = (
            torch.clamp(torch.round(w / scales) + zeros, min_int, max_int) - zeros
        ) * scales
    assert torch.isnan(w).sum() == 0

    w = w.reshape(org_w_shape)

    if get_scale_zp:
        return w, scales.view(w.shape[0], -1), zeros.view(w.shape[0], -1)
    else:
        return w


# weight quantization
@torch.no_grad()
def auto_clip_layer(
    w, input_feat, n_bit, q_config, n_grid=20, max_shrink=0.5, n_sample_token=512
):
    assert w.dim() == 2
    org_w_shape = w.shape
    # w           [co, ci]      -> [co, 1, n_group, group size]
    # input_feat  [n_token, ci] -> [1, n_token, n_group, group size]
    group_size = (
        q_config["q_group_size"] if q_config["q_group_size"] > 0 else w.shape[1]
    )
    input_feat = input_feat.view(-1, input_feat.shape[-1])
    input_feat = input_feat.reshape(1, input_feat.shape[0], -1, group_size)
    input_feat = input_feat[:, 0 :: input_feat.shape[1] // n_sample_token]
    w = w.reshape(w.shape[0], 1, -1, group_size)

    oc_batch_size = 256 if w.shape[0] % 256 == 0 else 64  # prevent OOM
    assert w.shape[0] % oc_batch_size == 0
    w_all = w
    best_max_val_all = []

    for i_b in range(w.shape[0] // oc_batch_size):
        w = w_all[i_b * oc_batch_size : (i_b + 1) * oc_batch_size]

        org_max_val = w.abs().amax(dim=-1, keepdim=True)  # co, 1, n_group, 1

        best_max_val = org_max_val.clone()
        min_errs = torch.ones_like(org_max_val) * 1e9
        input_feat = input_feat.to(w.device)
        org_out = (input_feat * w).sum(dim=-1)  # co, n_token, n_group

        for i_s in range(int(max_shrink * n_grid)):
            max_val = org_max_val * (1 - i_s / n_grid)
            min_val = -max_val
            cur_w = torch.clamp(w, min_val, max_val)
            q_w = pseudo_quantize_tensor(cur_w, n_bit=n_bit, **q_config)
            cur_out = (input_feat * q_w).sum(dim=-1)

            # co, 1, n_group, 1
            err = (cur_out - org_out).pow(2).mean(dim=1).view(min_errs.shape)
            del cur_w
            del cur_out
            cur_best_idx = err < min_errs
            min_errs[cur_best_idx] = err[cur_best_idx]
            best_max_val[cur_best_idx] = max_val[cur_best_idx]
        best_max_val_all.append(best_max_val)

    best_max_val = torch.cat(best_max_val_all, dim=0)

    del input_feat
    del org_out
    gc.collect()
    torch.cuda.empty_cache()
    return best_max_val.squeeze(1)


@torch.no_grad()
def auto_clip_block(module, w_bit, q_config, input_feat):
    named_linears = {
        name: m for name, m in module.named_modules() if isinstance(m, nn.Linear)
    }

    clip_list = []
    for name in named_linears:
        # due to qk bmm, it is hard to clip precisely
        if any([_ in name for _ in ["q_", "k_", "query", "key", "Wqkv"]]):
            continue
        named_linears[name].cuda()
        max_val = auto_clip_layer(
            named_linears[name].weight, input_feat[name], n_bit=w_bit, q_config=q_config
        )
        clip_list.append((name, max_val))
        named_linears[name].cpu()
    return clip_list


@torch.no_grad()
def apply_clip(module, clip_list):
    def get_op_by_name(module, op_name):
        for name, m in module.named_modules():
            if name == op_name:
                return m
        raise ValueError(f"Cannot find op {op_name} in module {module}")

    for name, max_val in clip_list:
        layer = get_op_by_name(module, name)
        layer.cuda()
        max_val = max_val.to(layer.weight.device)
        org_shape = layer.weight.shape
        layer.weight.data = layer.weight.data.reshape(*max_val.shape[:2], -1)
        layer.weight.data = torch.clamp(layer.weight.data, -max_val, max_val)
        layer.weight.data = layer.weight.data.reshape(org_shape)
        layer.cpu()



import gc
import torch
import torch.nn as nn

from transformers.models.llama.modeling_llama import LlamaDecoderLayer, LlamaRMSNorm
from transformers.activations import GELUActivation

from .qmodule import (
    ScaledActivation,
    get_op_by_name,
    get_op_name,
    set_op_by_name,
    append_str_prefix,
    pseudo_quantize_tensor,
)
from ..tensor_parallel import ColumnParallelLinear, RowParallelLinear


__all__ = ["auto_scale_block", "apply_scale"]


@torch.no_grad()
def get_weight_scale(weight, q_group_size=-1):
    org_shape = weight.shape
    if q_group_size > 0:
        weight = weight.view(-1, q_group_size)
    scale = weight.abs() / weight.abs().amax(dim=1, keepdim=True)
    scale = scale.view(org_shape)
    scale = scale.mean(0)
    return scale


@torch.no_grad()
def get_act_scale(x):
    return x.abs().view(-1, x.shape[-1]).mean(0)


@torch.no_grad()
def scale_ln_fcs(ln, fcs, scales):
    if not isinstance(fcs, list):
        fcs = [fcs]

    scales = scales.to(ln.weight.device)

    ln.weight.div_(scales)
    if hasattr(ln, "bias") and ln.bias is not None:
        ln.bias.div_(scales)

    for fc in fcs:
        fc.weight.mul_(scales.view(1, -1))

    for p in ln.parameters():
        assert torch.isnan(p).sum() == 0
    for fc in fcs:
        for p in fc.parameters():
            assert torch.isnan(p).sum() == 0


@torch.no_grad()
def scale_fc_fc(fc1, fc2, scales):
    assert isinstance(fc1, (nn.Linear, ColumnParallelLinear))
    assert isinstance(fc2, (nn.Linear, RowParallelLinear))
    # assert fc1.out_features == fc2.in_features

    scales = scales.to(fc1.weight.device)

    # fc1.weight.div_(scales.view(-1, 1))
    fc1.weight[-scales.size(0) :].div_(scales.view(-1, 1))
    if fc1.bias is not None:
        fc1.bias.div_(scales.view(-1))

    fc2.weight.mul_(scales.view(1, -1))

    for p in fc1.parameters():
        assert torch.isnan(p).sum() == 0
    for p in fc2.parameters():
        assert torch.isnan(p).sum() == 0


@torch.no_grad()
def auto_scale_block(module, module_kwargs, w_bit, q_config, input_feat):

    if w_bit is not None:

        def w_quantize_func(p):
            return pseudo_quantize_tensor(
                p,
                n_bit=w_bit,
                **q_config,
            ).detach()

    else:

        def w_quantize_func(p):
            return p

    if "use_cache" in module_kwargs:
        module_kwargs.pop("use_cache")

    # find the best scale ratio
    def _search_module_scale(block, linears2scale: list, x, kwargs={}):
        # w: co, ci
        # x: n, ci
        x = x.to(next(block.parameters()).device)
        with torch.no_grad():
            org_out = block(x, **kwargs)
            if isinstance(org_out, tuple):
                org_out = org_out[0]
        x_max = get_act_scale(x)

        best_error = float("inf")
        best_ratio = -1
        best_scales = None

        n_grid = 20
        history = []

        org_sd = {k: v.cpu() for k, v in block.state_dict().items()}
        for ratio in range(n_grid):
            ratio = ratio * 1 / n_grid
            scales = x_max.pow(ratio).clamp(min=1e-4).view(-1)
            scales = scales / (scales.max() * scales.min()).sqrt()
            for fc in linears2scale:
                fc.weight.mul_(scales.view(1, -1).to(fc.weight.device))
                fc.weight.data = w_quantize_func(fc.weight.data) / (scales.view(1, -1))
            out = block(x, **kwargs)
            if isinstance(out, tuple):
                out = out[0]

            loss = (
                (org_out - out).float().pow(2).mean().item()
            )  # float prevents overflow
            history.append(loss)
            is_best = loss < best_error
            if is_best:
                best_error = loss
                best_ratio = ratio
                best_scales = scales
            block.load_state_dict(org_sd)
        if best_ratio == -1:
            raise Exception
        best_scales = best_scales.view(-1)

        assert torch.isnan(best_scales).sum() == 0, best_scales
        return best_scales.detach()

    def _auto_get_scale(prev_op, layers, inp, module2inspect=None, kwargs={}):
        # module2inspect: if given, we will check the output diff of this module instead of layers
        if module2inspect is None:
            assert len(layers) == 1
            module2inspect = layers[0]

        scales = _search_module_scale(module2inspect, layers, inp, kwargs)
        scales = scales.detach().cpu()
        # prev_op_name, [layer_name], scale
        return (
            get_op_name(module, prev_op),
            tuple([get_op_name(module, m) for m in layers]),
            scales,
        )

    scales_list = []  # return the searched scales

    if isinstance(module, LlamaDecoderLayer):
        # attention input
        scales_list.append(
            _auto_get_scale(
                prev_op=module.input_layernorm,
                layers=[
                    module.self_attn.q_proj,
                    module.self_attn.k_proj,
                    module.self_attn.v_proj,
                ],
                inp=input_feat["self_attn.q_proj"],
                module2inspect=module.self_attn,
                kwargs=module_kwargs,
            )
        )
        # attn out
        # Please refer to https://github.com/mit-han-lab/llm-awq/pull/67#issue-1850622696
        if module.self_attn.v_proj.weight.shape == module.self_attn.o_proj.weight.shape:
            scales_list.append(
                _auto_get_scale(
                    prev_op=module.self_attn.v_proj,
                    layers=[module.self_attn.o_proj],
                    inp=input_feat["self_attn.o_proj"],
                )
            )
        # fc1
        scales_list.append(
            _auto_get_scale(
                prev_op=module.post_attention_layernorm,
                layers=[module.mlp.gate_proj, module.mlp.up_proj],
                inp=input_feat["mlp.gate_proj"],
                module2inspect=module.mlp,
            )
        )
        # fc2
        scales_list.append(
            _auto_get_scale(
                prev_op=module.mlp.up_proj,
                layers=[module.mlp.down_proj],
                inp=input_feat["mlp.down_proj"],
            )
        )

    else:
        raise NotImplementedError(f"{type(module)} not supported yet!")

    return scales_list


def apply_scale(module, scales_list, input_feat_dict=None):
    for prev_op_name, layer_names, scales in scales_list:
        prev_op = get_op_by_name(module, prev_op_name)
        layers = [get_op_by_name(module, name) for name in layer_names]

        prev_op.cuda()
        for layer in layers:
            layer.cuda()
        scales.cuda()

        from ..models.model import RMSNorm

        if isinstance(prev_op, (nn.Linear, ColumnParallelLinear)):
            assert len(layers) == 1
            scale_fc_fc(prev_op, layers[0], scales)
        elif isinstance(prev_op, (nn.LayerNorm, LlamaRMSNorm, RMSNorm)):
            scale_ln_fcs(prev_op, layers, scales)
        else:
            raise NotImplementedError(f"prev_op {type(prev_op)} not supported yet!")

        # apply the scaling to input feat if given; prepare it for clipping
        if input_feat_dict is not None:
            for layer_name in layer_names:
                inp = input_feat_dict[layer_name]
                inp.div_(scales.view(1, -1).to(inp.device))

        prev_op.cpu()
        for layer in layers:
            layer.cpu()
        scales.cpu()



import torch
import torch.nn as nn
from EETQ import w8_a16_gemm


class WeightOnlyLinear(nn.Module):
    def __init__(self, in_features, out_features, bias=None, name=None):
        super().__init__()

        self.in_features = in_features
        self.out_features = out_features

        self.register_buffer(
            "q_weight",
            torch.empty(
                (in_features, out_features), dtype=torch.int8, requires_grad=False
            ),
        )
        self.register_buffer(
            "q_scale_col",
            torch.empty((out_features), dtype=torch.float16, requires_grad=False),
        )

        if bias:

            self.register_buffer(
                "bias",
                torch.empty((out_features), dtype=torch.float16, requires_grad=False),
            )
        else:
            self.bias = None

        self.name = name

    @torch.no_grad()
    def forward(self, x):

        shape = x.shape[:-1] + (self.out_features,)

        inputs = x.reshape(-1, x.shape[-1])
        M = inputs.shape[0]

        y = w8_a16_gemm(inputs, self.q_weight, self.q_scale_col)

        if self.bias is not None:
            y += self.bias
        return y.reshape(shape)



"""
This file has adaption of open-source code from the following sources:
- The sampling implementation is originally from SGLang
  (https://github.com/sgl-project/sglang/blob/main/python/sglang/srt/layers/sampler.py),
  licensed under Apache 2.0.
"""

from logging import getLogger
from typing import Any, Tuple

import numpy as np
import torch

from chitu.global_vars import get_global_args

logger = getLogger(__name__)


def try_import_opt_dep(pkg_name: str, opt_dep_name: str) -> Tuple[Any, bool]:
    try:
        return __import__(pkg_name), True
    except ImportError:

        class ReportErrorWhenUsed:
            def __getattr__(self, item):
                raise ImportError(
                    f"Optional dependency '{opt_dep_name}' is not installed. "
                    f"Please refer to README.md for installation instructions."
                )

        return ReportErrorWhenUsed(), False


def is_layer(layer_name, full_name):
    return (
        f".{layer_name}." in full_name
        or full_name.startswith(layer_name + ".")
        or full_name.endswith("." + layer_name)
    )


def compute_layer_dist_in_pipe(num_layers, world_size):
    args = get_global_args()
    if args.infer.pp_layer_partition is not None:
        assert (
            len(args.infer.pp_layer_partition) == world_size
            and sum(args.infer.pp_layer_partition) == args.models.n_layers
        ), f"pp_layer_partition must be a list of length {world_size} and sum up to {args.models.n_layers}"
        num_layers_of_each_rank = args.infer.pp_layer_partition
    else:
        num_layers_of_each_rank = [
            num_layers // world_size + (1 if i < num_layers % world_size else 0)
            for i in range(world_size)
        ]
        # If non-divisible, make the fisrst and the last rank to have fewer layers, because they have pre-layers and post-layers
        if world_size > 2 and num_layers_of_each_rank[0] > num_layers_of_each_rank[-2]:
            num_layers_of_each_rank[0] -= 1
            num_layers_of_each_rank[-2] += 1
    return num_layers_of_each_rank


def top_k_top_p_min_p_sampling_from_probs_torch(
    probs: torch.Tensor,
    top_ks: torch.Tensor,
    top_ps: torch.Tensor,
    min_ps: torch.Tensor = None,  # TODO support min_ps
):
    """A top-k, top-p and min-p sampling implementation with native pytorch operations."""
    probs_sort, probs_idx = probs.sort(dim=-1, descending=True)
    probs_sum = torch.cumsum(probs_sort, dim=-1)
    # min_p_thresholds = probs_sort[:, 0] * min_ps
    probs_sort[(probs_sum - probs_sort) > top_ps.view(-1, 1)] = 0.0
    probs_sort[
        torch.arange(0, probs.shape[-1], device=probs.device).view(1, -1)
        >= top_ks.view(-1, 1)
    ] = 0.0
    # probs_sort[probs_sort < min_p_thresholds.view(-1, 1)] = 0.0
    probs_sort.div_(probs_sort.max(dim=-1, keepdim=True)[0])
    sampled_index = torch.multinomial(probs_sort, num_samples=1)
    batch_next_token_ids = torch.gather(probs_idx, dim=1, index=sampled_index).view(-1)
    return batch_next_token_ids


class VarLens:
    def __init__(self, tokens, device) -> None:
        self.lens = torch.tensor(
            [len(t) for t in tokens], device=device, dtype=torch.int32
        )
        self.cpu_prefix_lens = [0]
        for t in tokens:
            self.cpu_prefix_lens.append(self.cpu_prefix_lens[-1] + len(t))
        self.prefix_lens = torch.tensor(
            self.cpu_prefix_lens, device=device, dtype=torch.int32
        )
        self.cpu_lens = [len(t) for t in tokens]
        self.max_len = int(torch.max(self.lens))
        self.total_len = int(torch.sum(self.lens))
        self.position_ids = torch.from_numpy(
            np.concatenate([np.arange(l) for l in self.cpu_lens])
        ).to(device)


def get_config_dir_path():
    # Deprecated, but we need to support Python 3.8. importlib.resources is preferred in the future.
    import pkg_resources

    return pkg_resources.resource_filename("chitu", "config")



from typing import List, Optional

import torch
import torch.nn.functional as F
from torch import nn

from chitu.attn_backend import AttnBackend
from chitu.models.model import Attention, RMSNorm, Transformer, TransformerBlock
from chitu.tensor_parallel import (
    ColumnParallelLinear,
    RowParallelLinear,
    VocabParallelEmbedding,
    get_tp_size,
)


class AttentionLlama(Attention):
    def __init__(self, args, layer_id, cache, attn_backend):
        super().__init__(layer_id, cache, attn_backend)
        self.n_kv_heads = args.n_heads if args.n_kv_heads is None else args.n_kv_heads
        model_parallel_size = get_tp_size()
        self.n_local_heads = args.n_heads // model_parallel_size
        self.n_local_kv_heads = self.n_kv_heads // model_parallel_size
        self.n_rep = self.n_local_heads // self.n_local_kv_heads
        self.head_dim = args.dim // args.n_heads

        self.wq = ColumnParallelLinear(
            args.dim,
            args.n_heads * self.head_dim,
            has_bias=False,
            gather_output=False,
        )
        self.wk = ColumnParallelLinear(
            args.dim,
            self.n_kv_heads * self.head_dim,
            has_bias=False,
            gather_output=False,
        )
        self.wv = ColumnParallelLinear(
            args.dim,
            self.n_kv_heads * self.head_dim,
            has_bias=False,
            gather_output=False,
        )
        self.wo = RowParallelLinear(
            args.n_heads * self.head_dim,
            args.dim,
            has_bias=False,
            input_is_parallel=True,
        )

    def _run_linear(self, x):
        return self.wq(x), self.wk(x), self.wv(x)

    def _run_output_linear(self, x):
        return self.wo(x)


class TransformerLlama(Transformer):
    def __init__(
        self,
        params,
        cache,
        *,
        max_position_embeddings: int,
        pipeline_parallel_size: int,
        model_parallel_size: int,
        attn_backend: AttnBackend,
        op_impl: str,
        merge_qkv_gate_up: bool = False,
        **kvargs,
    ):
        super().__init__(
            params,
            cache,
            max_position_embeddings=max_position_embeddings,
            pipeline_parallel_size=pipeline_parallel_size,
            model_parallel_size=model_parallel_size,
            attn_backend=attn_backend,
            op_impl=op_impl,
            **kvargs,
        )
        self.op_impl = op_impl
        if merge_qkv_gate_up:
            raise NotImplementedError("merge_qkv_gate_up is not supported in llama")

    def _get_tensor_column_parallel_layer_names(self) -> List[str]:
        return ["wq", "wk", "wv", "w1", "w3", "output", "embed"]

    def _get_tensor_row_parallel_layer_names(self) -> List[str]:
        return ["wo", "w2"]

    def _get_pre_layer_prefixes(self) -> List[str]:
        return ["tok_embeddings."]

    def _get_post_layer_prefixes(self) -> List[str]:
        return ["output.", "norm."]

    def _get_layer_i_prefixes(self, i: int) -> List[str]:
        return [f"layers.{i}."]

    def _init_pre_layers(self):
        self.tok_embeddings = VocabParallelEmbedding(
            self.params.vocab_size, self.params.dim
        )

    def _init_layers(self, cache, attn_backend, op_impl):
        self.layers = torch.nn.ModuleList()
        for layer_id in range(self.local_begin_layer_id, self.local_end_layer_id):
            self.layers.append(
                TransformerBlockLlama(
                    layer_id, self.params, cache, attn_backend, self.op_impl
                )
            )

    def _init_post_layers(self):
        self.norm = RMSNorm(self.params.dim, eps=self.params.norm_eps)
        self.output = ColumnParallelLinear(
            self.params.dim, self.params.vocab_size, has_bias=False
        )

    def _pre_layers(self, h):
        return self.tok_embeddings(h)

    def _post_layers(self, h):
        """NOTE: _post_layers is assumed to be a token-wise computation"""
        h = self.norm(h)
        h = self.output(h)
        return h


class FeedForwardLlama(nn.Module):
    def __init__(
        self,
        dim: int,
        hidden_dim: int,
        multiple_of: int,
        ffn_dim_multiplier: Optional[float],
    ):
        super().__init__()
        hidden_dim = int(2 * hidden_dim / 3)
        # custom dim factor multiplier
        if ffn_dim_multiplier is not None:
            hidden_dim = int(ffn_dim_multiplier * hidden_dim)
        hidden_dim = multiple_of * ((hidden_dim + multiple_of - 1) // multiple_of)

        self.w1 = ColumnParallelLinear(
            dim, hidden_dim, has_bias=False, gather_output=False
        )
        self.w2 = RowParallelLinear(
            hidden_dim, dim, has_bias=False, input_is_parallel=True
        )
        self.w3 = ColumnParallelLinear(
            dim, hidden_dim, has_bias=False, gather_output=False
        )

    def forward(self, x):
        return self.w2(F.silu(self.w1(x)) * self.w3(x))


class TransformerBlockLlama(TransformerBlock):
    def __init__(self, layer_id: int, args, cache, attn_backend, op_impl):
        super().__init__(layer_id, args, cache, attn_backend, op_impl)
        self.attention = AttentionLlama(args, layer_id, cache, attn_backend)
        self.feed_forward = FeedForwardLlama(
            dim=args.dim,
            hidden_dim=4 * args.dim,
            multiple_of=args.multiple_of,
            ffn_dim_multiplier=args.ffn_dim_multiplier,
        )
        self.attention_norm = RMSNorm(args.dim, eps=args.norm_eps)
        self.ffn_norm = RMSNorm(args.dim, eps=args.norm_eps)

    def forward(
        self,
        x: torch.Tensor,
        freqs_cis_cos: torch.Tensor,
        freqs_cis_sin: torch.Tensor,
        varlens=None,
    ):
        h = self.attention(
            self.attention_norm(x), freqs_cis_cos, freqs_cis_sin, varlens
        )
        h += x
        out = h + self.feed_forward(self.ffn_norm(h))
        return out



import itertools
import math
import os
from dataclasses import dataclass
from logging import getLogger
from pathlib import Path
from typing import Any, List, Mapping, Optional, Tuple

import numpy as np
import torch
import torch.distributed
import torch.nn.functional as F
from torch import nn

from chitu.attn_backend import AttnBackend
from chitu.cache_manager import PagedKVCacheManager
from chitu.global_vars import get_global_args, get_timers, set_global_variables
from chitu.muxi_utils import has_tbsgemm, tbsgemm
from chitu.ops import apply_rotary_pos_emb
from chitu.tensor_parallel import get_tp_group, get_tp_rank
from chitu.tokenizer import ChatFormat, ChatFormatHF, Tokenizer, TokenizerHF
from chitu.utils import VarLens, compute_layer_dist_in_pipe, is_layer
from chitu.cuda_graph import make_dispatched_graphed_callables
from chitu.device_type import is_muxi, get_device_name

logger = getLogger(__name__)


class RMSNorm(nn.Module):
    """
    Root Mean Square Layer Normalization (RMSNorm).

    Args:
        dim (int): Dimension of the input tensor.
        eps (float): Epsilon value for numerical stability. Defaults to 1e-6.
    """

    def __init__(self, dim: int, eps: float = 1e-6):
        super().__init__()
        self.dim = dim
        self.eps = eps
        self.weight = nn.Parameter(torch.ones(dim))

    def _naive_norm(self, x, compute_dtype):
        dtype = x.dtype
        x = x.to(compute_dtype)
        y = x * torch.rsqrt(x.pow(2).mean(-1, keepdim=True) + self.eps)
        return y.to(dtype) * self.weight

    def forward(self, x: torch.Tensor, compute_dtype=None):
        """
        Forward pass for RMSNorm.

        Args:
            x (torch.Tensor): Input tensor.
            compute_dtype (torch.dtype, optional): The dtype to use for computation. Defaults to the
                dtype of the input tensor.

        Returns:
            torch.Tensor: Normalized tensor with the same shape as input.
        """
        # NOTE: Although F.rms_norm uses different dtypes inside itself, and some models directly
        # pass float16 tensors to it, our CI shows it does not work for some models, especially GPTQ
        # quantized models. Maybe we should make the dtype optional.

        if has_tbsgemm and x.dtype == torch.float16:
            output = tbsgemm.norm(x, self.weight)
            return output
        else:
            if compute_dtype is None:
                compute_dtype = torch.float32
            if hasattr(F, "rms_norm"):
                dtype = x.dtype
                return F.rms_norm(
                    x.to(compute_dtype), (self.dim,), self.weight, self.eps
                ).to(dtype)
            else:  # Old PyTorch versions
                return self._naive_norm(x, compute_dtype=compute_dtype)


def precompute_freqs_cis(dim: int, end: int, theta: float = 10000.0, device=None):
    freqs = 1.0 / (
        theta ** (torch.arange(0, dim, 2, device=device)[: (dim // 2)].float() / dim)
    )
    t = torch.arange(end, device=device, dtype=torch.float32)
    freqs = torch.outer(t, freqs)
    freqs_cis = torch.polar(torch.ones_like(freqs), freqs)  # complex64
    return freqs_cis


class Attention(nn.Module):
    def __init__(self, layer_id, cache, attn_backend):
        super().__init__()
        self.layer_id = layer_id
        self.cache = cache
        self.attn_backend = attn_backend

    def _run_linear(self, x):
        raise NotImplementedError

    def _run_output_linear(self, x):
        raise NotImplementedError

    def prefill_forward(
        self,
        x: torch.Tensor,
        freqs_cis_cos: torch.Tensor,
        freqs_cis_sin: torch.Tensor,
        varlens,
    ):
        bs_seq, _ = x.shape
        xq, xk, xv = self._run_linear(x)
        xq = xq.view(bs_seq, self.n_local_heads, self.head_dim)
        xk = xk.view(bs_seq, self.n_local_kv_heads, self.head_dim)
        xv = xv.view(bs_seq, self.n_local_kv_heads, self.head_dim)
        xq, xk = apply_rotary_pos_emb(
            xq, xk, freqs_cis_cos, freqs_cis_sin, rotary_type="llama"
        )
        self.cache.finalize_cache_bylayer_prefill(
            xk, xv, self.cache.curr_req_ids, self.cache.curr_varlens, self.layer_id
        )
        output = self.attn_backend.attn_varlen_func(
            xq,
            xk,
            xv,
            varlens.prefix_lens,
            varlens.prefix_lens,
            varlens.max_len,
            varlens.max_len,
            causal=True,
        ).view(bs_seq, -1)
        return self._run_output_linear(output)

    def decode_forward(
        self, x: torch.Tensor, freqs_cis_cos: torch.Tensor, freqs_cis_sin: torch.Tensor
    ):
        bsz, seqlen, _ = x.shape
        assert seqlen == 1, "decode_forward only supports single token decoding"
        xq, xk, xv = self._run_linear(x)

        xq = xq.view(-1, self.n_local_heads, self.head_dim)
        xk = xk.view(-1, self.n_local_kv_heads, self.head_dim)
        xv = xv.view(-1, self.n_local_kv_heads, self.head_dim)

        xq, xk = apply_rotary_pos_emb(
            xq, xk, freqs_cis_cos, freqs_cis_sin, rotary_type="llama"
        )

        xq = xq.view(bsz, seqlen, self.n_local_heads, self.head_dim)
        xk = xk.view(bsz, seqlen, self.n_local_kv_heads, self.head_dim)
        xv = xv.view(bsz, seqlen, self.n_local_kv_heads, self.head_dim)

        cache = self.cache.get_cache_decode(self.layer_id)
        cache_k = cache[0]
        cache_v = cache[1]
        cache_seqlens_excl_this_decode = self.cache.get_gpu_seq_lens_excl_this_decode()
        output = self.attn_backend.attn_with_kvcache(
            xq,
            cache_k,
            cache_v,
            xk,
            xv,
            cache_seqlens=cache_seqlens_excl_this_decode,
        ).view(bsz, seqlen, -1)
        return self._run_output_linear(output)

    def decode_forward_paged(
        self, x: torch.Tensor, freqs_cis_cos: torch.Tensor, freqs_cis_sin: torch.Tensor
    ):
        bsz, seqlen, _ = x.shape
        assert seqlen == 1, "decode_forward only supports single token decoding"
        xq, xk, xv = self._run_linear(x)

        xq = xq.view(-1, self.n_local_heads, self.head_dim)
        xk = xk.view(-1, self.n_local_kv_heads, self.head_dim)
        xv = xv.view(-1, self.n_local_kv_heads, self.head_dim)

        xq, xk = apply_rotary_pos_emb(
            xq, xk, freqs_cis_cos, freqs_cis_sin, rotary_type="llama"
        )

        xq = xq.view(bsz, seqlen, self.n_local_heads, self.head_dim)
        xk = xk.view(bsz, seqlen, self.n_local_kv_heads, self.head_dim)
        xv = xv.view(bsz, seqlen, self.n_local_kv_heads, self.head_dim)

        block_table = self.cache.get_gpu_block_table()
        cache_seqlens_excl_this_decode = self.cache.get_gpu_seq_lens_excl_this_decode()
        paged_k_cache, paged_v_cache = self.cache.get_paged_kv_cache(self.layer_id)
        output = self.attn_backend.attn_with_kvcache(
            xq,
            paged_k_cache,
            paged_v_cache,
            xk,
            xv,
            cache_seqlens=cache_seqlens_excl_this_decode,
            block_table=block_table,
        ).view(bsz, seqlen, -1)
        return self._run_output_linear(output)

    def forward(self, x, freqs_cis_cos, freqs_cis_sin, varlens=None):
        if varlens is not None:  # prefill
            return self.prefill_forward(x, freqs_cis_cos, freqs_cis_sin, varlens)
        elif isinstance(self.cache, PagedKVCacheManager):
            return self.decode_forward_paged(x, freqs_cis_cos, freqs_cis_sin)
        else:
            return self.decode_forward(x, freqs_cis_cos, freqs_cis_sin)


class TransformerBlock(nn.Module):

    def __init__(self, layer_id: int, args, cache, attn_backend, op_impl):
        super().__init__()
        self.n_heads = args.n_heads
        self.dim = args.dim
        self.head_dim = args.dim // args.n_heads
        self.layer_id = layer_id
        self.timers = get_timers()

    def forward(self):
        raise NotImplementedError


class Transformer(nn.Module):
    def __init__(
        self,
        params,
        cache,
        *,
        max_position_embeddings: int,
        pipeline_parallel_size: int,
        model_parallel_size: int,
        attn_backend: AttnBackend,
        op_impl: str,
        **kvargs,
    ):
        super().__init__()
        self.cache = cache
        self.attn_backend = attn_backend
        self.op_impl = op_impl
        self.rank = torch.distributed.get_rank()
        self.local_rank = int(os.environ.get("LOCAL_RANK", 0))
        self.world_size = torch.distributed.get_world_size()
        self.device = torch.device(self.local_rank)

        self.pipeline_parallel_size = pipeline_parallel_size
        self.model_parallel_size = model_parallel_size
        self.pipeline_exec = pipeline_parallel_size > 1
        self.tensor_exec = model_parallel_size > 1

        self.tp_size = model_parallel_size
        self.pp_size = pipeline_parallel_size
        self.pp_stage = self.rank // self.model_parallel_size
        self.pp_main_rank = (self.rank // model_parallel_size) * model_parallel_size
        self.pp_end_stage = (self.world_size - 1) // model_parallel_size
        self.tp_group = get_tp_group()

        self.params = params
        self.vocab_size = params.vocab_size
        self.global_n_layers = params.n_layers

        if self.pipeline_exec:
            num_layers_of_each_rank = compute_layer_dist_in_pipe(
                self.global_n_layers, self.pipeline_parallel_size
            )
            first_layer_id_of_each_rank = list(
                itertools.accumulate([0] + num_layers_of_each_rank)
            )
            self.local_begin_layer_id = first_layer_id_of_each_rank[self.pp_stage]
            self.local_end_layer_id = first_layer_id_of_each_rank[self.pp_stage + 1]
        else:
            self.local_begin_layer_id = 0
            self.local_end_layer_id = self.global_n_layers

        if not self.pipeline_exec or self.pp_stage == 0:
            self._init_pre_layers()
        self._init_layers(cache, attn_backend=attn_backend, op_impl=op_impl)
        if not self.pipeline_exec or self.pp_stage == self.pipeline_parallel_size - 1:
            self._init_post_layers()

        self.precompute_freqs_cis(max_position_embeddings, self.device)

    def _get_tensor_column_parallel_layer_names(self) -> List[str]:
        raise NotImplementedError

    def _get_tensor_row_parallel_layer_names(self) -> List[str]:
        raise NotImplementedError

    def _get_pre_layer_prefixes(self) -> List[str]:
        raise NotImplementedError

    def _get_post_layer_prefixes(self) -> List[str]:
        raise NotImplementedError

    def _get_layer_i_prefixes(self, i: int) -> List[str]:
        raise NotImplementedError

    def _chunk_checkpoint_for_pipeline_parallel(
        self,
        checkpoint,
        num_layers: int,
        rank: int,
        world_size: int,
    ):
        keys = checkpoint.keys()
        partial_checkpoint = {}

        num_layers_of_each_rank = compute_layer_dist_in_pipe(num_layers, world_size)
        first_layer_id_of_each_rank = list(
            itertools.accumulate([0] + num_layers_of_each_rank)
        )

        for i in range(
            first_layer_id_of_each_rank[rank], first_layer_id_of_each_rank[rank + 1]
        ):
            for key in keys:
                if i == 0:
                    for prefix in self._get_pre_layer_prefixes():
                        if key.startswith(prefix):
                            partial_checkpoint[key] = checkpoint[key]
                for prefix in self._get_layer_i_prefixes(i):
                    if key.startswith(prefix):
                        local_i = i - first_layer_id_of_each_rank[rank]
                        partial_checkpoint[
                            key.replace(f"layers.{i}.", f"layers.{local_i}.", 1)
                        ] = checkpoint[key]
                if i == num_layers - 1:
                    for prefix in self._get_post_layer_prefixes():
                        if key.startswith(prefix):
                            partial_checkpoint[key] = checkpoint[key]
        return partial_checkpoint

    def _chunk_checkpoint_for_tensor_parallel(
        self,
        checkpoint,
        rank: int,
        world_size: int,
    ):
        keys = checkpoint.keys()
        partial_checkpoint = {}

        cpl_names = self._get_tensor_column_parallel_layer_names()
        rpl_names = self._get_tensor_row_parallel_layer_names()

        for name, param in checkpoint.items():
            if any(is_layer(s, name) for s in cpl_names):
                if name.endswith("weight") or (
                    self.params.type == "deepseek-v3" and name.endswith("scale")
                ):
                    chunks = torch.chunk(param, world_size, dim=0)
                    partial_checkpoint[name] = chunks[rank]
                elif name.endswith("bias"):
                    chunks = torch.chunk(param, world_size, dim=-1)
                    partial_checkpoint[name] = chunks[rank]
                else:
                    assert False, f"Illegal parallel tensor {name}"
            elif any(is_layer(s, name) for s in rpl_names):
                if name.endswith("weight") or (
                    self.params.type == "deepseek-v3" and name.endswith("scale")
                ):
                    chunks = torch.chunk(param, world_size, dim=1)
                    partial_checkpoint[name] = chunks[rank]
                elif name.endswith("bias"):
                    # Rank 0 needs a full bias and only rank 0 needs it
                    if get_tp_rank() == 0:
                        partial_checkpoint[name] = param
                else:
                    assert False, f"Illegal parallel tensor {name}"
            else:
                partial_checkpoint[name] = param
        return partial_checkpoint

    def load_state_dict_parallel(
        self,
        state_dict: Mapping[str, Any],
        skip_preprocess: bool = False,
        *args,
        **kwargs,
    ):
        if not skip_preprocess:
            if self.pipeline_exec:
                state_dict = self._chunk_checkpoint_for_pipeline_parallel(
                    state_dict, self.global_n_layers, self.pp_stage, self.pp_size
                )
            if self.tensor_exec:
                state_dict = self._chunk_checkpoint_for_tensor_parallel(
                    state_dict, self.rank % self.tp_size, self.tp_size
                )
        self.load_state_dict(
            state_dict, skip_preprocess=skip_preprocess, *args, **kwargs
        )

    def load_state_dict(
        self,
        state_dict: Mapping[str, Any],
        skip_preprocess: bool = False,
        *args,
        **kwargs,
    ):
        super().load_state_dict(state_dict, *args, **kwargs)

    def _init_pre_layers(self):
        raise NotImplementedError

    def _init_layers(self, cache, attn_backend):
        raise NotImplementedError

    def _init_post_layers(self):
        raise NotImplementedError

    def _pre_layers(self, h):
        raise NotImplementedError

    def _post_layers(self, h):
        """NOTE: _post_layers is assumed to be a token-wise computation"""
        raise NotImplementedError

    def precompute_freqs_cis(self, max_position_embeddings, device):
        self.freqs_cis = precompute_freqs_cis(
            self.params.dim // self.params.n_heads,
            max_position_embeddings * 2,
            self.params.rope_theta,
            device=device,
        )

    def prepare_freqs_cis_prefill(self, varlens):
        curr_freqs_cis = self.freqs_cis[self.cache.curr_varlens.position_ids]
        return curr_freqs_cis.real.contiguous(), curr_freqs_cis.imag.contiguous()

    def prepare_freqs_cis_decode(self):
        curr_freqs_cis = self.freqs_cis[self.cache.get_gpu_seq_lens_excl_this_decode()]
        return curr_freqs_cis.real.contiguous(), curr_freqs_cis.imag.contiguous()

    @torch.inference_mode()
    def prefill_single_device(self, tokens, varlens=None):
        if isinstance(
            tokens, list
        ):  # else use tensor variable passed by TensorExecutor
            varlens = VarLens(tokens, self.device)
            tokens = torch.from_numpy(np.concatenate(tokens)).to(self.device)
        freqs_cis_cos, freqs_cis_sin = self.prepare_freqs_cis_prefill(varlens)
        h = self._pre_layers(tokens)
        for it, layer in enumerate(self.layers):
            h = layer(h, freqs_cis_cos, freqs_cis_sin, varlens)
        tmp = varlens.cpu_prefix_lens[1:]
        h = h[[item - 1 for item in tmp]]
        h = self._post_layers(h)  # Exec post layers AFTER cutting the last token off
        h = h.float()
        return h

    @torch.inference_mode()
    def decode_single_device(self, tokens, freqs_cis_cos, freqs_cis_sin):
        h = self._pre_layers(tokens)
        for it, layer in enumerate(self.layers):
            h = layer(h, freqs_cis_cos, freqs_cis_sin)
        h = self._post_layers(h)
        h = h.float()
        return h

    @torch.inference_mode()
    def prefill_pipeline(self, tokens):

        varlens = self.cache.curr_varlens
        freqs_cis_cos, freqs_cis_sin = self.prepare_freqs_cis_prefill(varlens)

        # start of model
        if self.pp_stage == 0:
            h = self._pre_layers(tokens)
        else:
            h = tokens
        # layers
        for it, layer in enumerate(self.layers):
            h = layer(h, freqs_cis_cos, freqs_cis_sin, varlens)
        # end of model
        if self.pp_stage == self.pp_end_stage:
            tmp = varlens.cpu_prefix_lens[1:]
            h = h[[item - 1 for item in tmp]]
            h = self._post_layers(
                h
            )  # Exec post layers AFTER cutting the last token off

            h = h.float()

        return h

    @torch.inference_mode()
    def decode_pipeline(self, tokens, freqs_cis_cos, freqs_cis_sin):
        if self.pp_stage == 0:
            h = self._pre_layers(tokens)
        else:
            h = tokens
        for it, layer in enumerate(self.layers):
            h = layer(h, freqs_cis_cos, freqs_cis_sin)
        if self.pp_stage == self.pp_end_stage:
            h = self._post_layers(h)
            h = h.float()

        return h

    @torch.inference_mode()
    def prefill(self, tokens, varlens=None):
        if self.pipeline_exec:
            return self.prefill_pipeline(tokens)
        elif self.tensor_exec:
            return self.prefill_single_device(tokens, varlens)
        else:
            return self.prefill_single_device(tokens)

    def prepare_decoding_attn(self):
        cache_seqlens_excl_this_decode = self.cache.get_gpu_seq_lens_excl_this_decode()
        cache_seqlens_incl_this_decode = self.cache.get_gpu_seq_lens_incl_this_decode()
        block_table = self.cache.get_gpu_block_table()
        block_size = self.cache.get_block_size()
        self.attn_backend.prepare_metadata_for_decode(
            cache_seqlens_excl_this_decode,
            cache_seqlens_incl_this_decode,
            block_table,
            block_size,
        )

    @torch.inference_mode()
    def decode(self, tokens, seq_lens):
        self.prepare_decoding_attn()

        batch_size = len(seq_lens)
        max_batch_size = get_global_args().infer.max_reqs

        use_cuda_graph = get_global_args().infer.use_cuda_graph

        if use_cuda_graph:
            if get_global_args().infer.cache_type == "skew":
                raise NotImplementedError(
                    'CUDA graph is currently not supported for infer.cache_type="skew"'
                )

        @make_dispatched_graphed_callables(
            sample_args=(tokens,),
            sample_kwargs={},
            args_max_nelem=(tokens.numel() // batch_size * max_batch_size,),
            kwargs_max_nelem={},
            output_max_nelem_callback=lambda n: n // batch_size * max_batch_size,
            enable=use_cuda_graph,
        )
        def do_decode(tokens):
            freqs_cis_cos, freqs_cis_sin = self.prepare_freqs_cis_decode()
            if self.pipeline_exec:
                return self.decode_pipeline(tokens, freqs_cis_cos, freqs_cis_sin)
            else:
                return self.decode_single_device(tokens, freqs_cis_cos, freqs_cis_sin)

        return do_decode(batch_size, tokens)



import asyncio
import json
import threading
import time
from datetime import datetime
from logging import getLogger
from typing import List, Optional

from pydantic import BaseModel

from chitu.backend import Backend

logger = getLogger(__name__)


class ChatCompletionResponse(BaseModel):
    id: str
    choices: list
    usage: Optional[dict] = None


class AsyncDataStream:
    def __init__(self):
        self.tokenizer = Backend.tokenizer
        self.seqs: List[str] = []
        self.tokens_len: int = 0
        self.chars_len: int = 0
        self.cache_tokens: List[int] = []
        self.stop_signal = False
        self.lock = threading.Lock()
        self.data_event = asyncio.Event()
        self.is_reasoning = False
        self.reasoning_len = 0

    def add_data(self, value: int):
        with self.lock:
            if self.reasoning_handle(value):
                return
            self.tokens_len += 1
            self.cache_tokens.append(value)
            s = self.tokenizer.decode(self.cache_tokens)
            if "\ufffd" in s:
                return
            if not self.tokenizer.force_full_seq_decode:
                self.cache_tokens.clear()
                self.seqs.append(s)
                self.chars_len += len(s)
            else:
                self.seqs.append(s[self.chars_len :])
                self.chars_len = len(s)
        self.data_event.set()

    def send_stop_signal(self):
        with self.lock:
            self.stop_signal = True
        self.data_event.set()

    def reasoning_handle(self, value: int):
        rs_token_id = Backend.args.models.get("rs_token_id", -1)
        re_token_id = Backend.args.models.get("re_token_id", -1)
        if rs_token_id == -1 or re_token_id == -1:
            return False
        if not self.is_reasoning and self.tokens_len == 0 and value == rs_token_id:
            self.is_reasoning = True
            return True
        if self.is_reasoning and value == re_token_id:
            self.is_reasoning = False
            self.reasoning_len = len(self.seqs)
            return True

    def is_reasoning_content(self):
        return self.is_reasoning or self.index - 1 < self.reasoning_len

    def __aiter__(self):
        self.index = 0
        return self

    async def __anext__(self):
        while True:
            with self.lock:
                if self.stop_signal and self.index >= len(self.seqs):
                    raise StopAsyncIteration
                if self.index < len(self.seqs):
                    result = self.seqs[self.index]
                    self.index += 1
                    return result
            self.data_event.clear()
            await self.data_event.wait()


class AsyncResponse:
    def __init__(self, req):
        self.req = req
        self.id = req.request_id
        self.async_stream = req.async_stream

    def stream_generator(self):
        async def stream_response():
            async for data in self.async_stream:
                if data:
                    delta = {}
                    if self.async_stream.is_reasoning_content():
                        delta["reasoning_content"] = f"{data}"
                    else:
                        delta["content"] = f"{data}"
                    chunk = ChatCompletionResponse(
                        id=self.id,
                        choices=[
                            {
                                "index": 0,
                                "delta": delta,
                                "finish_reason": None,
                                "time_stamp": datetime.now().strftime("%H:%M:%S:%f"),
                            }
                        ],
                    )
                    data = chunk.model_dump_json(exclude_none=True)
                    yield f"data: {data}\n\n"

            chunk = ChatCompletionResponse(
                id=self.id,
                choices=[
                    {
                        "index": 0,
                        "delta": {"content": ""},
                        "finish_reason": self.req.finish_reason,
                    }
                ],
                usage={
                    "prompt_tokens": f"{self.req.prompt_len}",
                    "completion_tokens": f"{self.async_stream.tokens_len}",
                    "total_tokens": f"{self.async_stream.tokens_len + self.req.prompt_len}",
                },
            )
            data = chunk.model_dump_json(exclude_none=True)
            yield f"data: {data}\n\n"
            logger.debug(
                f"Completed_{self.id}: {self.req.output}, token_len: {self.async_stream.tokens_len}\n"
            )
            yield "data: [DONE]\n\n"

        return stream_response()

    async def full_generator(self):
        text = []
        async for data in self.async_stream:
            text.append(data)
        r_len = self.async_stream.reasoning_len
        message = {}
        message["role"] = "assistant"

        if r_len:
            message["reasoning_content"] = "".join(text[:r_len])
        message["content"] = "".join(text[r_len:])

        full_response = ChatCompletionResponse(
            id=self.id,
            choices=[{"index": 0, "message": message}],
            usage={
                "prompt_tokens": f"{self.req.prompt_len}",
                "completion_tokens": f"{self.async_stream.tokens_len}",
                "total_tokens": f"{self.async_stream.tokens_len + self.req.prompt_len}",
            },
        )
        logger.debug(
            f"Completed_{self.id}: {self.req.output}, token_len: {self.async_stream.tokens_len}\n"
        )
        return full_response



"""
This file has adaption of open-source code from the following sources:
- The interface class (AttnBackend) is originally from flash_attn (https://github.com/Dao-AILab/flash-attention), licensed under BSD-3-Clause.
- The implementation of the reference backend (RefAttnBackend) is originally from flash_attn's test (https://github.com/Dao-AILab/flash-attention/blob/main/tests/test_flash_attn.py), licensed under BSD-3-Clause.
"""

__all__ = ["AttnBackend", "FlashAttnBackend", "RefAttnBackend"]

import abc
import math
from typing import Optional, Union

import torch

from chitu.global_vars import get_global_args
from chitu.ops import append_to_paged_kv_cache
from chitu.triton_decode_attention import mla_decode
from chitu.utils import try_import_opt_dep
from chitu.static_tensor import StaticTensor

flash_attn, has_flash_attn = try_import_opt_dep("flash_attn", "flash_attn")
flash_mla, has_flash_mla = try_import_opt_dep("flash_mla", "flash_mla")
flashinfer, has_flashinfer = try_import_opt_dep("flashinfer", "flashinfer")


class AttnBackend(abc.ABC):
    """
    Interface class for all attention implementations
    """

    def prepare_metadata_for_decode(self, *args, **kwargs):
        pass

    def attn_varlen_func(
        self,
        q,
        k,
        v,
        cu_seqlens_q,
        cu_seqlens_k,
        max_seqlen_q,
        max_seqlen_k,
        dropout_p=0.0,
        causal=False,
        window_size=(-1, -1),  # -1 means infinite context window
        softcap=0.0,  # 0.0 means deactivated
        softmax_scale=None,
    ):
        """
        dropout_p should be set to 0.0 during evaluation.

        Supports multi-query and grouped-query attention (MQA/GQA) by passing in K, V with fewer heads
        than Q. Note that the number of heads in Q must be divisible by the number of heads in KV.
        For example, if Q has 6 heads and K, V have 2 heads, head 0, 1, 2 of Q will attention to head
        0 of K, V, and head 3, 4, 5 of Q will attention to head 1 of K, V.

        If causal=True, the causal mask is aligned to the bottom right corner of the attention matrix.
        For example, if seqlen_q = 2 and seqlen_k = 5, the causal mask (1 = keep, 0 = masked out) is:
            1 1 1 1 0
            1 1 1 1 1
        If seqlen_q = 5 and seqlen_k = 2, the causal mask is:
            0 0
            0 0
            0 0
            1 0
            1 1
        If the row of the mask is all zero, the output will be zero.

        If window_size != (-1, -1), implements sliding window local attention. Query at position i
        will only attend to keys between
        [i + seqlen_k - seqlen_q - window_size[0], i + seqlen_k - seqlen_q + window_size[1]] inclusive.

        Arguments:
            q: (total_q, nheads, headdim), where total_q = total number of query tokens in the batch.
            k: (total_k, nheads_k, headdim), where total_k = total number of key tokens in the batch.
            v: (total_k, nheads_k, headdim), where total_k = total number of key tokens in the batch.
            cu_seqlens_q: (batch_size + 1,), dtype torch.int32. The cumulative sequence lengths
               of the sequences in the batch, used to index into q.
            cu_seqlens_k: (batch_size + 1,), dtype torch.int32. The cumulative sequence lengths
               of the sequences in the batch, used to index into kv.
            max_seqlen_q: int. Maximum query sequence length in the batch.
            max_seqlen_k: int. Maximum key sequence length in the batch.
            dropout_p: float. Dropout probability.
            Default to 1 / sqrt(headdim).
            causal: bool. Whether to apply causal attention mask (e.g., for auto-regressive modeling).
            window_size: (left, right). If not (-1, -1), implements sliding window local attention.
            softcap: float. Anything > 0 activates softcapping attention.
            softmax_scale: float. The scaling of QK^T before applying softmax. Default to 1 / sqrt(headdim).
        Return:
            out: (total, nheads, headdim).
        """
        raise NotImplementedError()

    def attn_with_kvcache(
        self,
        q,
        k_cache,
        v_cache,
        k=None,
        v=None,
        cache_seqlens: Optional[Union[(int, torch.Tensor)]] = None,
        cache_leftpad: Optional[torch.Tensor] = None,
        block_table: Optional[torch.Tensor] = None,
        causal=False,
        window_size=(-1, -1),  # -1 means infinite context window
        softcap=0.0,  # 0.0 means deactivated
        softmax_scale=None,
    ):
        """
        If k and v are not None, k_cache and v_cache will be updated *inplace* with the new values from
        k and v. This is useful for incremental decoding: you can pass in the cached keys/values from
        the previous step, and update them with the new keys/values from the current step, and do
        attention with the updated cache, all in 1 kernel.

        If you pass in k / v, you must make sure that the cache is large enough to hold the new values.
        For example, the KV cache could be pre-allocated with the max sequence length, and you can use
        cache_seqlens to keep track of the current sequence lengths of each sequence in the batch.

        See tests/test_flash_attn.py::test_flash_attn_kvcache for examples of how to use this function.

        Supports multi-query and grouped-query attention (MQA/GQA) by passing in KV with fewer heads
        than Q. Note that the number of heads in Q must be divisible by the number of heads in KV.
        For example, if Q has 6 heads and K, V have 2 heads, head 0, 1, 2 of Q will attention to head
        0 of K, V, and head 3, 4, 5 of Q will attention to head 1 of K, V.

        If causal=True, the causal mask is aligned to the bottom right corner of the attention matrix.
        For example, if seqlen_q = 2 and seqlen_k = 5, the causal mask (1 = keep, 0 = masked out) is:
            1 1 1 1 0
            1 1 1 1 1
        If seqlen_q = 5 and seqlen_k = 2, the causal mask is:
            0 0
            0 0
            0 0
            1 0
            1 1
        If the row of the mask is all zero, the output will be zero.

        If window_size != (-1, -1), implements sliding window local attention. Query at position i
        will only attend to keys between
        [i + seqlen_k - seqlen_q - window_size[0], i + seqlen_k - seqlen_q + window_size[1]] inclusive.

        Note: Does not support backward pass.

        Arguments:
            q: (batch_size, seqlen, nheads, headdim)
            k_cache: (batch_size_cache, seqlen_cache, nheads_k, headdim) if there's no block_table,
                or (num_blocks, page_block_size, nheads_k, headdim) if there's a block_table (i.e. paged KV cache)
                page_block_size must be a multiple of 256.
            v_cache: (batch_size_cache, seqlen_cache, nheads_k, headdim) if there's no block_table,
                or (num_blocks, page_block_size, nheads_k, headdim) if there's a block_table (i.e. paged KV cache)
            k [optional]: (batch_size, seqlen_new, nheads_k, headdim). If not None, we concatenate
                k with k_cache, starting at the indices specified by cache_seqlens.
            v [optional]: (batch_size, seqlen_new, nheads_k, headdim). Similar to k.
            cache_seqlens: int, or (batch_size,), dtype torch.int32. The sequence lengths of the
                KV cache.
            cache_leftpad: (batch_size,), dtype torch.int32. The index that the KV cache starts. If None, assume 0.
            block_table [optional]: (batch_size, max_num_blocks_per_seq), dtype torch.int32.
            causal: bool. Whether to apply causal attention mask (e.g., for auto-regressive modeling).
            window_size: (left, right). If not (-1, -1), implements sliding window local attention.
            softcap: float. Anything > 0 activates softcapping attention.
            softmax_scale: float. The scaling of QK^T before applying softmax. Default to 1 / sqrt(headdim).

        Return:
            out: (batch_size, seqlen, nheads, headdim).
        """
        raise NotImplementedError()


class FlashAttnBackend(AttnBackend):

    def __init__(self):
        super().__init__()

    def attn_varlen_func(
        self,
        q,
        k,
        v,
        cu_seqlens_q,
        cu_seqlens_k,
        max_seqlen_q,
        max_seqlen_k,
        dropout_p=0.0,
        causal=False,
        window_size=(-1, -1),  # -1 means infinite context window
        softcap=0.0,  # 0.0 means deactivated
        softmax_scale=None,
    ):
        # These are arguments only accpeted by new enough flash_attn,
        # so don't pass them if they are set to default values
        extra_kvargs = {}
        if softcap != 0.0:
            extra_kvargs["softcap"] = softcap

        return flash_attn.flash_attn_varlen_func(
            q,
            k,
            v,
            cu_seqlens_q,
            cu_seqlens_k,
            max_seqlen_q,
            max_seqlen_k,
            dropout_p=dropout_p,
            causal=causal,
            window_size=window_size,
            softmax_scale=softmax_scale,
            **extra_kvargs,
        )

    def attn_with_kvcache(
        self,
        q,
        k_cache,
        v_cache,
        k=None,
        v=None,
        cache_seqlens: Optional[Union[(int, torch.Tensor)]] = None,
        cache_leftpad: Optional[torch.Tensor] = None,
        block_table: Optional[torch.Tensor] = None,
        causal=False,
        window_size=(-1, -1),  # -1 means infinite context window
        softcap=0.0,  # 0.0 means deactivated
        softmax_scale=None,
    ):
        # These are arguments only accpeted by new enough flash_attn,
        # so don't pass them if they are set to default values
        extra_kvargs = {}
        if cache_leftpad is not None:
            extra_kvargs["cache_leftpad"] = cache_leftpad
        if softcap != 0.0:
            extra_kvargs["softcap"] = softcap

        return flash_attn.flash_attn_with_kvcache(
            q,
            k_cache,
            v_cache,
            k=k,
            v=v,
            cache_seqlens=cache_seqlens,
            block_table=block_table,
            causal=causal,
            window_size=window_size,
            softmax_scale=softmax_scale,
            **extra_kvargs,
        )


class RefAttnBackend(AttnBackend):

    def __init__(self):
        super().__init__()

        import einops as _einops

        self._einops = _einops

    def _construct_local_mask(
        self,
        seqlen_q,
        seqlen_k,
        window_size=(-1, -1),  # -1 means infinite window size
        query_padding_mask=None,
        key_padding_mask=None,
        device=None,
        key_leftpad=None,
    ):
        row_idx = self._einops.rearrange(
            torch.arange(seqlen_q, device=device, dtype=torch.long), "s -> s 1"
        )
        col_idx = torch.arange(seqlen_k, device=device, dtype=torch.long)
        if key_leftpad is not None:
            key_leftpad = self._einops.rearrange(key_leftpad, "b -> b 1 1 1")
            col_idx = self._einops.repeat(
                col_idx, "s -> b 1 1 s", b=key_leftpad.shape[0]
            )
            col_idx = torch.where(col_idx >= key_leftpad, col_idx - key_leftpad, 2**32)
        sk = (
            seqlen_k
            if key_padding_mask is None
            else self._einops.rearrange(key_padding_mask.sum(-1), "b -> b 1 1 1")
        )
        sq = (
            seqlen_q
            if query_padding_mask is None
            else self._einops.rearrange(query_padding_mask.sum(-1), "b -> b 1 1 1")
        )
        if window_size[0] < 0:
            return col_idx > row_idx + sk - sq + window_size[1]
        else:
            sk = torch.full_like(col_idx, seqlen_k) if key_padding_mask is None else sk
            return torch.logical_or(
                col_idx > torch.minimum(row_idx + sk - sq + window_size[1], sk),
                col_idx < row_idx + sk - sq - window_size[0],
            )

    def _attention(
        self,
        q,
        k,
        v,
        query_padding_mask=None,
        key_padding_mask=None,
        attn_bias=None,
        dropout_p=0.0,
        dropout_mask=None,
        causal=False,
        window_size=(-1, -1),  # -1 means infinite window size
        softcap=0.0,
        upcast=True,
        reorder_ops=False,
        key_leftpad=None,
        softmax_scale=None,
    ):
        """
        Arguments:
            q: (batch_size, seqlen_q, nheads, head_dim)
            k: (batch_size, seqlen_k, nheads_k, head_dim)
            v: (batch_size, seqlen_k, nheads_k, head_dim)
            query_padding_mask: (batch_size, seqlen_q)
            key_padding_mask: (batch_size, seqlen_k)
            attn_bias: broadcastable to (batch_size, nheads, seqlen_q, seqlen_k)
            dropout_p: float
            dropout_mask: (batch_size, nheads, seqlen_q, seqlen_k)
            causal: whether to apply causal masking
            window_size: (int, int), left and right window size
            upcast: whether to cast all inputs to fp32, do all computation in fp32, then cast
                output back to fp16/bf16.
            reorder_ops: whether to change the order of operations (scaling k instead of scaling q, etc.)
                without changing the math. This is to estimate the numerical error from operation
                reordering.
        Output:
            output: (batch_size, seqlen_q, nheads, head_dim)
            attention: (batch_size, nheads, seqlen_q, seqlen_k), softmax after dropout
        """
        if causal:
            window_size = (window_size[0], 0)
        dtype_og = q.dtype
        if upcast:
            q, k, v = q.float(), k.float(), v.float()
        seqlen_q, seqlen_k = q.shape[1], k.shape[1]
        k = self._einops.repeat(k, "b s h d -> b s (h g) d", g=q.shape[2] // k.shape[2])
        v = self._einops.repeat(v, "b s h d -> b s (h g) d", g=q.shape[2] // v.shape[2])
        d = q.shape[-1]
        if softmax_scale is None:
            softmax_scale = 1.0 / math.sqrt(d)
        if not reorder_ops:
            scores = torch.einsum("bthd,bshd->bhts", q * softmax_scale, k)
        else:
            scores = torch.einsum("bthd,bshd->bhts", q, k * softmax_scale)
        if softcap > 0:
            scores = scores / softcap
            scores = scores.tanh()
            scores = scores * softcap
        if key_padding_mask is not None:
            scores.masked_fill_(
                self._einops.rearrange(~key_padding_mask, "b s -> b 1 1 s"),
                float("-inf"),
            )
        if window_size[0] >= 0 or window_size[1] >= 0:
            local_mask = self._construct_local_mask(
                seqlen_q,
                seqlen_k,
                window_size,
                query_padding_mask,
                key_padding_mask,
                q.device,
                key_leftpad=key_leftpad,
            )
            scores.masked_fill_(local_mask, float("-inf"))
        if attn_bias is not None:
            scores = scores + attn_bias
        attention = torch.softmax(scores, dim=-1).to(v.dtype)
        # Some rows might be completely masked out so we fill them with zero instead of NaN
        if window_size[0] >= 0 or window_size[1] >= 0:
            attention = attention.masked_fill(
                torch.all(local_mask, dim=-1, keepdim=True), 0.0
            )
        # We want to mask here so that the attention matrix doesn't have any NaNs
        # Otherwise we'll get NaN in dV
        if query_padding_mask is not None:
            attention = attention.masked_fill(
                self._einops.rearrange(~query_padding_mask, "b s -> b 1 s 1"), 0.0
            )
        dropout_scaling = 1.0 / (1 - dropout_p)
        if dropout_mask is not None:
            attention_drop = attention.masked_fill(~dropout_mask, 0.0)
        else:
            attention_drop = attention
        output = torch.einsum("bhts,bshd->bthd", attention_drop, v * dropout_scaling)
        if query_padding_mask is not None:
            output.masked_fill_(
                self._einops.rearrange(~query_padding_mask, "b s -> b s 1 1"), 0.0
            )
        return output.to(dtype=dtype_og), attention.to(dtype=dtype_og)

    def attn_varlen_func(
        self,
        q,
        k,
        v,
        cu_seqlens_q,
        cu_seqlens_k,
        max_seqlen_q,
        max_seqlen_k,
        dropout_p=0.0,
        causal=False,
        window_size=(-1, -1),  # -1 means infinite context window
        softcap=0.0,  # 0.0 means deactivated
        softmax_scale=None,
    ):
        q_batch = torch.zeros(
            (cu_seqlens_q.shape[0] - 1,) + tuple(q.shape),
            dtype=q.dtype,
            device=q.device,
        )
        k_batch = torch.zeros(
            (cu_seqlens_k.shape[0] - 1,) + tuple(k.shape),
            dtype=k.dtype,
            device=k.device,
        )
        v_batch = torch.zeros(
            (cu_seqlens_k.shape[0] - 1,) + tuple(v.shape),
            dtype=v.dtype,
            device=v.device,
        )
        for i in range(cu_seqlens_q.shape[0] - 1):
            q_batch[i, 0 : cu_seqlens_q[i + 1] - cu_seqlens_q[i]] = q[
                cu_seqlens_q[i] : cu_seqlens_q[i + 1]
            ]
            k_batch[i, 0 : cu_seqlens_k[i + 1] - cu_seqlens_k[i]] = k[
                cu_seqlens_k[i] : cu_seqlens_k[i + 1]
            ]
            v_batch[i, 0 : cu_seqlens_k[i + 1] - cu_seqlens_k[i]] = v[
                cu_seqlens_k[i] : cu_seqlens_k[i + 1]
            ]
        output_batch, _ = self._attention(
            q_batch,
            k_batch,
            v_batch,
            dropout_p=dropout_p,
            causal=causal,
            window_size=window_size,
            softcap=softcap,
            softmax_scale=softmax_scale,
        )
        output = torch.empty(
            (cu_seqlens_q[-1] - cu_seqlens_q[0],) + output_batch.shape[2:],
            dtype=output_batch[0].dtype,
            device=output_batch[0].device,
        )
        for i in range(cu_seqlens_q.shape[0] - 1):
            # fmt: off
            output[
                cu_seqlens_q[i] - cu_seqlens_q[0] :
                cu_seqlens_q[i + 1] - cu_seqlens_q[0]
            ] = output_batch[i, 0 : cu_seqlens_q[i + 1] - cu_seqlens_q[i]]
            # fmt: on
        return output

    def attn_with_kvcache(
        self,
        q,
        k_cache,
        v_cache,
        k=None,
        v=None,
        cache_seqlens: Optional[Union[(int, torch.Tensor)]] = None,
        cache_leftpad: Optional[torch.Tensor] = None,
        block_table: Optional[torch.Tensor] = None,
        causal=False,
        window_size=(-1, -1),  # -1 means infinite context window
        softcap=0.0,  # 0.0 means deactivated
        softmax_scale=None,
    ):
        assert block_table is None, "block_table is not supported in RefAttnBackend"
        if cache_seqlens is int or cache_seqlens.ndim == 0:
            cache_seqlens = torch.full(
                (q.shape[0],), cache_seqlens, dtype=torch.long, device=q.device
            )
        arange = self._einops.rearrange(
            torch.arange(k_cache.shape[1], device=k_cache.device), "s -> 1 s"
        )
        cache_seqlens_expanded = self._einops.rearrange(cache_seqlens, "b -> b 1")
        if k is None and q is None:
            key_padding_mask = arange < cache_seqlens_expanded
        elif k is not None and q is not None:
            key_padding_mask = arange < cache_seqlens_expanded + 1
            for i in range(cache_seqlens.shape[0]):
                k_cache[i][cache_seqlens[i]] = k[i]
                v_cache[i][cache_seqlens[i]] = v[i]
        output, _ = self._attention(
            q,
            k_cache,
            v_cache,
            None,
            key_padding_mask,
            causal=causal,
            window_size=window_size,
            softcap=softcap,
            key_leftpad=cache_leftpad,
            softmax_scale=softmax_scale,
        )
        return output


class FlashMLABackend(RefAttnBackend):
    def __init__(self):
        super().__init__()

        self.args = get_global_args()
        self.mtp_size = 1
        self.kv_heads = 1
        self.local_n_heads = self.args.models.n_heads // self.args.infer.tp_size
        self.metadata = None
        self.num_splits = None

    def prepare_metadata_for_decode(
        self,
        cache_seqlens_excl_this_decode,
        cache_seqlens_incl_this_decode,
        block_table,
        block_size,
        softmax_scale=None,
    ):
        max_batch_size = get_global_args().infer.max_reqs
        metadata, num_splits = flash_mla.get_mla_metadata(
            cache_seqlens_incl_this_decode,
            self.mtp_size * self.local_n_heads // self.kv_heads,
            self.kv_heads,
        )
        if self.metadata is None:
            self.metadata = StaticTensor(metadata)  # `metadata` has a fixed shape
        else:
            self.metadata.set(metadata)
        if self.num_splits is None:
            self.num_splits = StaticTensor(
                num_splits, max_nelem=max_batch_size + 1
            )  # `num_splits`'s shape is always (batch_size + 1,)
        else:
            self.num_splits.set(num_splits)

    def mla_attn_with_kvcache(
        self,
        q_nope,
        q_pe,
        kv_cache,
        kv,
        cache_seqlens_excl_this_decode: Union[(int, torch.Tensor)],
        cache_seqlens_incl_this_decode: Union[(int, torch.Tensor)],
        block_table: torch.Tensor,
        causal=True,
        window_size=(-1, -1),  # -1 means infinite context window
        softcap=0.0,  # 0.0 means deactivated
        softmax_scale=None,
    ):
        bsz = cache_seqlens_excl_this_decode.shape[0]

        q_nope_pe = torch.cat([q_nope, q_pe], dim=-1)
        q_nope_pe = q_nope_pe.view(bsz, 1, q_nope_pe.shape[-2], q_nope_pe.shape[-1])

        append_to_paged_kv_cache(
            kv_cache, block_table, kv, cache_seqlens_excl_this_decode
        )

        kv_cache = kv_cache.unsqueeze(2)

        output, _ = flash_mla.flash_mla_with_kvcache(
            q_nope_pe,
            kv_cache,
            block_table,
            cache_seqlens_incl_this_decode,
            512,  # dv
            self.metadata.get(),
            self.num_splits.get(),
            causal=causal,
            softmax_scale=softmax_scale,
        )
        return output


class FlashInferBackend(RefAttnBackend):
    def __init__(self, tot_num_blocks):
        super().__init__()

        args = get_global_args()

        # FlashInfer accepts block tables for Q and KV in CSR format.
        # - For Q, it is trivial because the length for each sample is 1.
        # - For KV, we need to convert `block_table` to CSR format.
        # These buffers must be allocated when initializing
        # `flashinfer.mla.BatchMLAPagedAttentionWrapper` when cuda graph is enabled
        max_batch_size = get_global_args().infer.max_reqs
        self.q_indptr = StaticTensor(
            torch.empty(max_batch_size + 1, dtype=torch.int32, device="cuda")
        )
        self.kv_indptr = StaticTensor(
            torch.empty(max_batch_size + 1, dtype=torch.int32, device="cuda")
        )
        self.kv_indices = StaticTensor(
            torch.empty(tot_num_blocks, dtype=torch.int32, device="cuda")
        )
        self.seqlens = StaticTensor(
            torch.empty(max_batch_size, dtype=torch.int32, device="cuda")
        )

        self.mla_wrapper = flashinfer.mla.BatchMLAPagedAttentionWrapper(
            torch.empty(128 * 1024 * 1024, dtype=torch.int8).cuda(),
            use_cuda_graph=args.infer.use_cuda_graph,
            qo_indptr=self.q_indptr.get(),
            kv_indptr=self.kv_indptr.get(),
            kv_indices=self.kv_indices.get(),
            kv_len_arr=self.seqlens.get(),
            backend="auto",
        )

        self.local_n_heads = args.models.n_heads // args.infer.tp_size
        self.kv_lora_rank = args.models.kv_lora_rank
        self.qk_rope_head_dim = args.models.qk_rope_head_dim
        self.qk_nope_head_dim = args.models.qk_nope_head_dim

    def prepare_metadata_for_decode(
        self,
        cache_seqlens_excl_this_decode,
        cache_seqlens_incl_this_decode,
        block_table,
        block_size,
        softmax_scale=None,
    ):
        batch_size = cache_seqlens_incl_this_decode.shape[0]
        self.q_indptr.set(torch.arange(0, batch_size + 1).cuda().to(torch.int32))
        kv_indptr_list = []
        kv_indices_list = []
        tot_len = 0
        for i in range(batch_size):
            kv_indptr_list.append(tot_len)
            cur_len = (cache_seqlens_incl_this_decode[i].item() - 1) // block_size + 1
            kv_indices_list.append(block_table[i, :cur_len])
            tot_len += cur_len
        kv_indptr_list.append(tot_len)
        self.kv_indptr.set(torch.tensor(kv_indptr_list).cuda().to(torch.int32))
        self.kv_indices.set(torch.cat(kv_indices_list).cuda().to(torch.int32))
        self.seqlens.set(cache_seqlens_incl_this_decode)

        if softmax_scale is None:
            softmax_scale = 1.0 / (
                (self.qk_rope_head_dim + self.qk_nope_head_dim) ** 0.5
            )

        # Currently `self.mla_wrapper` holds fixed reserved buffers for CUDA graph, whose
        # sizes cannot be changed for different batch size. We have to forcely override
        # their shapes here.
        self.mla_wrapper._qo_indptr_buf = self.q_indptr.get()
        self.mla_wrapper._kv_indptr_buf = self.kv_indptr.get()
        self.mla_wrapper._kv_indices_buf = self.kv_indices.get()
        self.mla_wrapper._kv_len_arr_buf = self.seqlens.get()

        self.mla_wrapper.plan(
            self.q_indptr.get(),
            self.kv_indptr.get(),
            self.kv_indices.get(),
            self.seqlens.get(),
            num_heads=self.local_n_heads,
            head_dim_ckv=self.kv_lora_rank,
            head_dim_kpe=self.qk_rope_head_dim,
            page_size=block_size,
            causal=True,
            sm_scale=softmax_scale,
            q_data_type=torch.get_default_dtype(),
            kv_data_type=torch.get_default_dtype(),
        )

    def mla_attn_with_kvcache(
        self,
        q_nope,
        q_pe,
        kv_cache,
        kv,
        cache_seqlens_excl_this_decode: Union[(int, torch.Tensor)],
        cache_seqlens_incl_this_decode: Union[(int, torch.Tensor)],
        block_table: torch.Tensor,
        causal=True,
        window_size=(-1, -1),  # -1 means infinite context window
        softcap=0.0,  # 0.0 means deactivated
        softmax_scale=None,
    ):
        append_to_paged_kv_cache(
            kv_cache, block_table, kv, cache_seqlens_excl_this_decode
        )

        return self.mla_wrapper.run(
            q_nope,
            q_pe,
            kv_cache[..., : self.kv_lora_rank],
            kv_cache[..., self.kv_lora_rank :],
            return_lse=False,
        ).view(cache_seqlens_excl_this_decode.shape[0], 1, self.local_n_heads, -1)


class TritonAttnBackend(RefAttnBackend):
    def __init__(self):
        super().__init__()

        self.args = get_global_args()
        self.local_n_heads = self.args.models.n_heads // self.args.infer.tp_size
        self.kv_lora_rank = self.args.models.kv_lora_rank
        self.qk_rope_head_dim = self.args.models.qk_rope_head_dim
        self.qk_nope_head_dim = self.args.models.qk_nope_head_dim

    def prepare_metadata_for_decode(
        self,
        cache_seqlens_excl_this_decode,
        cache_seqlens_incl_this_decode,
        block_table,
        block_size,
        softmax_scale=None,
    ):
        self.block_size = block_size

    def mla_attn_with_kvcache(
        self,
        q_nope,
        q_pe,
        kv_cache,
        kv,
        cache_seqlens_excl_this_decode: Union[(int, torch.Tensor)],
        cache_seqlens_incl_this_decode: Union[(int, torch.Tensor)],
        block_table: torch.Tensor,
        causal=True,
        window_size=(-1, -1),  # -1 means infinite context window
        softcap=0.0,  # 0.0 means deactivated
        softmax_scale=None,
    ):
        append_to_paged_kv_cache(
            kv_cache, block_table, kv, cache_seqlens_excl_this_decode
        )

        B = q_nope.shape[0]

        o = torch.zeros(
            B,
            self.local_n_heads,
            self.kv_lora_rank,
            dtype=q_nope.dtype,
            device=q_nope.device,
        )

        num_kv_splits = 4

        attn_logits = torch.empty(
            (
                B,
                self.local_n_heads,
                num_kv_splits,
                self.kv_lora_rank + 1,
            ),
            dtype=torch.float32,
            device=q_nope.device,
        )

        assert kv_cache.ndim == 3  # (num_blocks, block_size, dim)
        kv_c_and_k_pe_cache = kv_cache
        kv_c_cache = kv_c_and_k_pe_cache[..., : self.kv_lora_rank]
        k_pe_cache = kv_c_and_k_pe_cache[..., self.kv_lora_rank :]

        PAGE_SIZE = kv_c_and_k_pe_cache.size(1)

        if softmax_scale is None:
            softmax_scale = (
                1.0 / ((self.qk_rope_head_dim + self.qk_nope_head_dim) ** 0.5),
            )

        mla_decode(
            q_nope,
            q_pe,
            kv_c_cache,
            k_pe_cache,
            o,
            block_table,
            cache_seqlens_incl_this_decode,
            attn_logits,
            num_kv_splits,
            softmax_scale,
            PAGE_SIZE,
        )

        return o.view(B, 1, self.local_n_heads, -1)



from typing import List

from chitu.utils import try_import_opt_dep

muxi_layout_kernels, has_muxi_layout_kernels = try_import_opt_dep(
    "muxi_layout_kernels", "muxi_layout_kernels"
)
tbsgemm, has_tbsgemm = try_import_opt_dep("tbsgemm", "muxi_w8a8_kernels")


def preprocess_weights_for_native_layout(
    checkpoint, rpl_names: List[str], cpl_names: List[str]
):
    def is_rpl_weight(key):
        for name in rpl_names:
            if key.endswith(f".{name}.weight"):
                return True
        return False

    def is_cpl_weight(key):
        for name in cpl_names:
            if key.endswith(f".{name}.weight"):
                return True
        return False

    new_checkpoint = {}
    for key in checkpoint.keys():
        if is_rpl_weight(key):  # Row parallel
            m, k = checkpoint[key].shape
            assert m % 64 == 0
            assert k % 128 == 0
            new_checkpoint[key] = (
                checkpoint[key]
                .reshape(m // 16, 16, k // 8, 8)
                .permute(0, 2, 1, 3)
                .contiguous()
                .reshape(m, k)  # Reshape back to for compatibility
            )
        elif is_cpl_weight(key):  # Column parallel
            m, k = checkpoint[key].shape
            assert m % 64 == 0
            assert k % 128 == 0
            new_checkpoint[key] = (
                checkpoint[key]
                .reshape(m // 16, 16, k // 8, 8)
                .permute(0, 2, 1, 3)
                .contiguous()
                .reshape(m, k)  # Reshape back to for compatibility
            )
        else:
            new_checkpoint[key] = checkpoint[key]
    return new_checkpoint


def linear_layout_contig_x_native_y(x, w, b=None):
    assert x.ndim == 2
    x_is_vector = x.shape[0] == 1
    if not x_is_vector:
        x_transposed = muxi_layout_kernels.layoutB(x)
    # w has already been transposed, but reshaped back for compatibility. We only need to "view" it again.
    w_transposed = w.view(w.shape[0] // 16, w.shape[1] // 8, 16, 8)
    if x_is_vector:
        y = muxi_layout_kernels.gemv_layoutA(w_transposed, x, bias=b)
        # View as 5D to be compatible with "native layout" but make n's tile to be 1.
        y = y.view(y.shape[1] // 32, 1, 4, 1, 8)
    elif x_transposed.shape[1] * 16 > 256:
        y = muxi_layout_kernels.muxi_hgemm_layout(w_transposed, x_transposed, bias=b)
        y = muxi_layout_kernels.layoutB(y)
    else:
        y = muxi_layout_kernels.gemm_layoutABC(w_transposed, x_transposed, bias=b)
    return y


def linear_layout_native_x_contig_y(x_transposed, w, b=None):
    assert x_transposed.ndim == 5
    x_is_vector = x_transposed.shape[1] == 1 and x_transposed.shape[3] == 1
    # w has already been transposed, but reshaped back for compatibility. We only need to "view" it again.
    w_transposed = w.view(w.shape[0] // 16, w.shape[1] // 8, 16, 8)
    if x_is_vector:
        y = muxi_layout_kernels.gemv_layoutA(
            w_transposed, x_transposed.view(1, -1), bias=b
        )
    elif x_transposed.shape[1] * 16 > 256:
        y = muxi_layout_kernels.muxi_hgemm_layout(w_transposed, x_transposed, bias=b)
    else:
        y = muxi_layout_kernels.gemm_layoutAB_ContinuousC(
            w_transposed, x_transposed, bias=b
        )
    return y


def linear_layout_contig_x_contig_y(x, w, b=None):
    assert x.ndim == 2
    x_is_vector = x.shape[0] == 1
    if not x_is_vector:
        x_transposed = muxi_layout_kernels.layoutB(x)
    # w has already been transposed, but reshaped back for compatibility. We only need to "view" it again.
    w_transposed = w.view(w.shape[0] // 16, w.shape[1] // 8, 16, 8)
    if x_is_vector:
        y = muxi_layout_kernels.gemv_layoutA(w_transposed, x, bias=b)
    elif x_transposed.shape[1] * 16 > 256:
        y = muxi_layout_kernels.muxi_hgemm_layout(w_transposed, x_transposed, bias=b)
    else:
        y = muxi_layout_kernels.gemm_layoutAB_ContinuousC(
            w_transposed, x_transposed, bias=b
        )
    return y



"""
This awq module has adaption of open-source code from the following sources:
- llm-awq (https://github.com/mit-han-lab/llm-awq?tab=MIT-1-ov-file), licensed under MIT.
"""

from .auto_clip import *
from .auto_scale import *
from .qmodule import *



import os
from logging import getLogger
from pathlib import Path
from typing import (
    AbstractSet,
    cast,
    Collection,
    Dict,
    Iterator,
    List,
    Literal,
    Sequence,
    TypedDict,
    Union,
)

import tiktoken
from tiktoken.load import load_tiktoken_bpe
from transformers import AutoTokenizer


logger = getLogger(__name__)


Role = Literal["system", "user", "assistant"]


class Message(TypedDict):
    role: Role
    content: str


Dialog = Sequence[Message]


class Tokenizer:
    """
    Tokenizing and encoding/decoding text using the Tiktoken tokenizer.
    """

    special_tokens: Dict[str, int]

    num_reserved_special_tokens = 256

    pat_str = r"(?i:'s|'t|'re|'ve|'m|'ll|'d)|[^\r\n\p{L}\p{N}]?\p{L}+|\p{N}{1,3}| ?[^\s\p{L}\p{N}]+[\r\n]*|\s*[\r\n]+|\s+(?!\S)|\s+"  # noqa: E501

    def __init__(self, model_path: str, force_full_seq_decode: bool = False):
        """
        Initializes the Tokenizer with a Tiktoken model.

        Args:
            model_path (str): The path to the Tiktoken model file.
            force_full_seq_decode (bool): If true, every time a new token is generated,
                the output string will be re-decoded from all the generated tokens. If
                there are `n` tokens, it will take `O(n^2)` time to decode them all. This
                is requried by some models.
        """
        self.force_full_seq_decode = force_full_seq_decode

        assert os.path.isfile(model_path), model_path

        mergeable_ranks = load_tiktoken_bpe(model_path)
        num_base_tokens = len(mergeable_ranks)
        special_tokens = [
            "<|begin_of_text|>",
            "<|end_of_text|>",
            "<|reserved_special_token_0|>",
            "<|reserved_special_token_1|>",
            "<|reserved_special_token_2|>",
            "<|reserved_special_token_3|>",
            "<|start_header_id|>",
            "<|end_header_id|>",
            "<|reserved_special_token_4|>",
            "<|eot_id|>",  # end of turn
        ] + [
            f"<|reserved_special_token_{i}|>"
            for i in range(5, self.num_reserved_special_tokens - 5)
        ]
        self.tokens_cache = []
        self.special_tokens = {
            token: num_base_tokens + i for i, token in enumerate(special_tokens)
        }
        self.model = tiktoken.Encoding(
            name=Path(model_path).name,
            pat_str=self.pat_str,
            mergeable_ranks=mergeable_ranks,
            special_tokens=self.special_tokens,
        )
        logger.debug(f"Reloaded tiktoken model from {model_path}")

        self.n_words: int = self.model.n_vocab
        # BOS / EOS token IDs
        self.bos_id: int = self.special_tokens["<|begin_of_text|>"]
        self.eos_id: int = self.special_tokens["<|end_of_text|>"]
        self.pad_id: int = -1
        self.stop_tokens = {
            self.special_tokens["<|end_of_text|>"],
            self.special_tokens["<|eot_id|>"],
        }
        logger.debug(
            f"#words: {self.n_words} - BOS ID: {self.bos_id} - EOS ID: {self.eos_id}"
        )

    def encode(
        self,
        s: str,
        *,
        bos: bool,
        eos: bool,
        allowed_special: Union[Literal["all"], AbstractSet[str]] = set(),
        disallowed_special: Union[Literal["all"], Collection[str]] = (),
    ) -> List[int]:
        """
        Encodes a string into a list of token IDs.

        Args:
            s (str): The input string to be encoded.
            bos (bool): Whether to prepend the beginning-of-sequence token.
            eos (bool): Whether to append the end-of-sequence token.
            allowed_tokens ("all"|set[str]): allowed special tokens in string
            disallowed_tokens ("all"|set[str]): special tokens that raise an error when in string

        Returns:
            list[int]: A list of token IDs.

        By default, setting disallowed_special=() encodes a string by ignoring
        special tokens. Specifically:
        - Setting `disallowed_special` to () will cause all text corresponding
          to special tokens to be encoded as natural text (insteading of raising
          an error).
        - Setting `allowed_special` to "all" will treat all text corresponding
          to special tokens to be encoded as special tokens.
        """
        assert type(s) is str, type(s)

        # The tiktoken tokenizer can handle <=400k chars without
        # pyo3_runtime.PanicException.
        TIKTOKEN_MAX_ENCODE_CHARS = 400_000

        # https://github.com/openai/tiktoken/issues/195
        # Here we iterate over subsequences and split if we exceed the limit
        # of max consecutive non-whitespace or whitespace characters.
        MAX_NO_WHITESPACES_CHARS = 25_000

        substrs = (
            substr
            for i in range(0, len(s), TIKTOKEN_MAX_ENCODE_CHARS)
            for substr in self._split_whitespaces_or_nonwhitespaces(
                s[i : i + TIKTOKEN_MAX_ENCODE_CHARS], MAX_NO_WHITESPACES_CHARS
            )
        )
        t: List[int] = []
        for substr in substrs:
            t.extend(
                self.model.encode(
                    substr,
                    allowed_special=allowed_special,
                    disallowed_special=disallowed_special,
                )
            )
        if bos:
            t.insert(0, self.bos_id)
        if eos:
            t.append(self.eos_id)
        return t

    def decode(self, t: Sequence[int]) -> str:
        """
        Decodes a list of token IDs into a string.

        Args:
            t (List[int]): The list of token IDs to be decoded.

        Returns:
            str: The decoded string.
        """
        # Typecast is safe here. Tiktoken doesn't do anything list-related with the sequence.
        if len(t) == 1 and t[0] in self.stop_tokens:
            return ""
        return self.model.decode(cast(List[int], t))

    @staticmethod
    def _split_whitespaces_or_nonwhitespaces(
        s: str, max_consecutive_slice_len: int
    ) -> Iterator[str]:
        """
        Splits the string `s` so that each substring contains no more than `max_consecutive_slice_len`
        consecutive whitespaces or consecutive non-whitespaces.
        """
        current_slice_len = 0
        current_slice_is_space = s[0].isspace() if len(s) > 0 else False
        slice_start = 0

        for i in range(len(s)):
            is_now_space = s[i].isspace()

            if current_slice_is_space ^ is_now_space:
                current_slice_len = 1
                current_slice_is_space = is_now_space
            else:
                current_slice_len += 1
                if current_slice_len > max_consecutive_slice_len:
                    yield s[slice_start:i]
                    slice_start = i
                    current_slice_len = 1
        yield s[slice_start:]


class ChatFormat:
    def __init__(self, tokenizer: Tokenizer):
        self.tokenizer = tokenizer

    def encode_header(self, message: Message) -> List[int]:
        tokens = []
        tokens.append(self.tokenizer.special_tokens["<|start_header_id|>"])
        tokens.extend(self.tokenizer.encode(message["role"], bos=False, eos=False))
        tokens.append(self.tokenizer.special_tokens["<|end_header_id|>"])
        tokens.extend(self.tokenizer.encode("\n\n", bos=False, eos=False))
        return tokens

    def encode_message(self, message: Message) -> List[int]:
        tokens = self.encode_header(message)
        tokens.extend(
            self.tokenizer.encode(message["content"].strip(), bos=False, eos=False)
        )
        tokens.append(self.tokenizer.special_tokens["<|eot_id|>"])
        return tokens

    def encode_dialog_prompt(self, dialog: Dialog) -> List[int]:
        tokens = []
        tokens.append(self.tokenizer.special_tokens["<|begin_of_text|>"])
        for message in dialog:
            tokens.extend(self.encode_message(message))
        # Add the start of an assistant message for the model to complete.
        tokens.extend(self.encode_header({"role": "assistant", "content": ""}))
        return tokens


class TokenizerHF:
    def __init__(
        self,
        path: str,
        trust_remote_code: bool = False,
        force_full_seq_decode: bool = False,
    ):
        self.force_full_seq_decode = force_full_seq_decode
        self.tokens_cache = []
        self.model = AutoTokenizer.from_pretrained(
            path, trust_remote_code=trust_remote_code
        )
        # self.model = AutoTokenizer.from_pretrained("Qwen/Qwen2-7B-Instruct")
        # Qwen2 don't set bos but have <|im_start|>
        # all special tokens: <|endoftext|> <|im_start|> <|im_end|>
        if "qwen2" in path.lower():
            self.bos_id = self.model.convert_tokens_to_ids("<|im_start|>")
        else:
            self.bos_id = self.model.bos_token_id
        self.eos_id = self.model.eos_token_id
        self.pad_id = self.model.pad_token_id
        self.stop_tokens = self.model.eos_token_id
        self.n_words = self.model.vocab_size

    def encode(self, s: str, bos: bool, eos: bool) -> List[int]:
        t = self.model.encode(s, add_special_tokens=False)
        if bos and self.bos_id is not None:
            t.insert(0, self.bos_id)
        if eos and self.eos_id is not None:
            t.append(self.eos_id)
        return t

    def decode(self, t: Sequence[int], skip_special_tokens=True) -> str:
        return self.model.decode(t, skip_special_tokens=True)


class ChatFormatHF:
    def __init__(self, tokenizer: TokenizerHF):
        self.tokenizer = tokenizer

    def encode_header(self, message: Message) -> List[int]:  # ???
        tokens = []
        tokens.extend(self.tokenizer.encode(message["role"], bos=True, eos=False))
        tokens.extend(self.tokenizer.encode("\n", bos=False, eos=False))
        return tokens

    def encode_message(self, message: Message) -> List[int]:
        tokens = self.encode_header(message)
        tokens.extend(
            self.tokenizer.encode(message["content"].strip(), bos=False, eos=True)
        )
        tokens.extend(self.tokenizer.encode("\n", bos=False, eos=False))
        return tokens

    def encode_dialog_prompt(self, dialog: Dialog) -> List[int]:
        tokens = []
        for message in dialog:
            tokens.extend(self.encode_message(message))
        tokens.extend(self.encode_header({"role": "assistant", "content": ""}))
        return tokens



import math
import torch
import torch.nn as nn
import awq_inference_engine  # with CUDA kernels
import tqdm
import gc

from ..tensor_parallel import ColumnParallelLinear, RowParallelLinear


def get_op_by_name(module, op_name):
    # get the op by its name relative to the module
    for name, m in module.named_modules():
        if name == op_name:
            return m
    raise ValueError(f"Cannot find op {op_name} in module {module}")


def set_op_by_name(layer, name, new_module):
    levels = name.split(".")
    if len(levels) > 1:
        mod_ = layer
        for l_idx in range(len(levels) - 1):
            if levels[l_idx].isdigit():
                mod_ = mod_[int(levels[l_idx])]
            else:
                mod_ = getattr(mod_, levels[l_idx])
        setattr(mod_, levels[-1], new_module)
    else:
        setattr(layer, name, new_module)


def get_op_name(module, op):
    # get the name of the op relative to the module
    for name, m in module.named_modules():
        if m is op:
            return name
    raise ValueError(f"Cannot find op {op} in module {module}")


def append_str_prefix(x, prefix):
    if isinstance(x, str):
        return prefix + x
    elif isinstance(x, tuple):
        return tuple([append_str_prefix(y, prefix) for y in x])
    elif isinstance(x, list):
        return [append_str_prefix(y, prefix) for y in x]
    else:
        return x


def make_divisible(c, divisor):
    return (c + divisor - 1) // divisor


def calculate_zeros_width(in_features, group_size=128, pack_num=8):
    if group_size >= 128:
        size_multiplier = 1
    elif group_size == 64:
        size_multiplier = 2
    elif group_size == 32:
        size_multiplier = 4
    else:
        raise NotImplementedError

    base_width = make_divisible(in_features // group_size, pack_num)
    base_width = make_divisible(base_width, size_multiplier) * size_multiplier
    return base_width


def pack_intweight(unpacked_qweight, interleave, kstride):
    # unpacked_qweight: [N, K]
    N = unpacked_qweight.shape[0]
    K = unpacked_qweight.shape[1]

    Packed_Kernel = unpacked_qweight.cpu().numpy().reshape(N, K // 32, 32)
    # np.arange(32).reshape(4, 4, 2).transpose(1, 0, 2) => [0, 1, 8, 9, 16, 17, 24, 25, ...]
    Packed_Kernel = Packed_Kernel.reshape(N, K // 32, 4, 4, 2).transpose(0, 1, 3, 2, 4)
    Packed_Kernel = Packed_Kernel.reshape(N, K // 32, 32)

    # reorder each 8 weights for fast dequantization
    # [0, 1, 2, 3, 4, 5, 6, 7] => [0, 2, 4, 6, 1, 3, 5, 7]
    Packed_Kernel = Packed_Kernel.reshape(N, K // 32, 4, 8)
    Packed_Kernel = Packed_Kernel.reshape(N, K // 32, 4, 4, 2).transpose(0, 1, 2, 4, 3)
    Packed_Kernel = Packed_Kernel.reshape(N, K)

    # interleaving every four rows
    Packed_Kernel = Packed_Kernel.reshape(
        N // interleave, interleave, K // kstride, kstride
    )
    # N // 4, K // 64, 4, 64
    Packed_Kernel = Packed_Kernel.transpose(0, 2, 1, 3)
    Packed_Kernel = Packed_Kernel.reshape(
        N // interleave, K // kstride, kstride, interleave
    )
    # Packing -> (N // 4, K // 64, 64)
    Packed_Kernel = (
        Packed_Kernel[..., 0]
        | (Packed_Kernel[..., 1] << 4)
        | (Packed_Kernel[..., 2] << 8)
        | (Packed_Kernel[..., 3] << 12)
    )
    # reshape to (N // 4, K), FP16 format
    Packed_Kernel = Packed_Kernel.reshape(N // interleave, K)
    qweight = (
        torch.tensor(Packed_Kernel.astype("int16"))
        .to(unpacked_qweight.device)
        .contiguous()
    )
    return qweight


class ScaledActivation(nn.Module):
    def __init__(self, module, scales):
        super().__init__()
        self.act = module
        self.scales = nn.Parameter(scales.data)

    def forward(self, x):
        return self.act(x) / self.scales.view(1, 1, -1).to(x.device)


class WQLinear(nn.Module):
    def __init__(self, w_bit, group_size, in_features, out_features, bias, dev):
        super().__init__()

        if w_bit not in [4]:
            raise NotImplementedError("Only 4-bit are supported for now.")

        self.in_features = in_features
        self.out_features = out_features
        self.w_bit = w_bit
        self.group_size = group_size if group_size != -1 else in_features
        self.split_k_iters = 8
        self.interleave = 4
        # quick sanity check (make sure aligment)
        assert self.in_features % self.group_size == 0
        assert out_features % (32 // self.w_bit) == 0
        pack_num = 32 // self.w_bit
        int16_pack_num = 16 // self.w_bit

        assert out_features % (self.interleave) == 0
        self.register_buffer(
            "qweight",
            torch.zeros(
                (
                    out_features // self.interleave,
                    in_features // int16_pack_num * self.interleave,
                ),
                dtype=torch.int16,
                device=dev,
            ),
        )
        self.register_buffer(
            "scales",
            torch.zeros(
                (
                    calculate_zeros_width(in_features, self.group_size) * pack_num,
                    out_features,
                ),
                dtype=torch.float16,
                device=dev,
            ),
        )
        self.register_buffer(
            "scaled_zeros",
            torch.zeros(
                (
                    calculate_zeros_width(in_features, self.group_size) * pack_num,
                    out_features,
                ),
                dtype=torch.float16,
                device=dev,
            ),
        )

        if bias:
            self.register_buffer(
                "bias", torch.zeros((out_features), dtype=torch.float16, device=dev)
            )
        else:
            self.bias = None

    @classmethod
    def from_linear(
        cls, linear, w_bit, group_size, init_only=False, scales=None, zeros=None
    ):
        awq_linear = cls(
            w_bit,
            group_size,
            linear.in_features,
            linear.out_features,
            linear.bias is not None,
            linear.weight.device,
        )
        if init_only:  # just prepare for loading sd
            return awq_linear

        # need scales and zeros info for real quantization
        assert scales is not None and zeros is not None
        scale_zeros = zeros * scales

        pack_num = 32 // awq_linear.w_bit
        qscales = torch.zeros(
            (
                scales.shape[0],
                calculate_zeros_width(linear.in_features, group_size) * pack_num,
            ),
            dtype=torch.float16,
            device=scales.device,
        )
        qscales[:, : scales.shape[1]] = scales
        # awq_linear.scales = scales.clone().half()
        awq_linear.scales = qscales.transpose(1, 0).contiguous()
        if linear.bias is not None:
            awq_linear.bias = linear.bias.clone().half()

        intweight = []
        for idx in range(awq_linear.in_features):
            intweight.append(
                torch.round(
                    (linear.weight.data[:, idx] + scale_zeros[:, idx // group_size])
                    / qscales[:, idx // group_size]
                ).to(torch.int)[:, None]
            )
        intweight = torch.cat(intweight, dim=1)
        # intweight = intweight.t().contiguous()
        intweight = intweight.to(dtype=torch.int32)
        awq_linear.qweight = pack_intweight(
            intweight.contiguous(), interleave=4, kstride=64
        )

        zeros = zeros.to(dtype=torch.int32)
        scaled_zeros = torch.zeros_like(qscales)
        # scaled_zeros[:, :scales.shape[1]] = -(qscales[:, :scales.shape[1]] * (zeros.to(torch.float32) - 8.0)).to(torch.float16)
        scaled_zeros[:, : scales.shape[1]] = -(
            qscales[:, : scales.shape[1]] * (zeros.to(torch.float32))
        ).to(torch.float16)
        awq_linear.scaled_zeros = scaled_zeros.transpose(1, 0).contiguous()

        return awq_linear

    @torch.no_grad()
    def forward(self, x):
        # out_shape = x.shape[:-1] + (self.out_features,)
        # inputs = x.reshape(-1, x.shape[-1])
        inputs = x
        # if inputs.numel() / inputs.shape[-1] < 8:
        # batch_size, n_tokens, _ = inputs.shape

        if inputs.shape[0] < 5:
            out = awq_inference_engine.gemv_forward_cuda_new(
                inputs,
                self.qweight,
                self.scales,
                self.scaled_zeros,
                inputs.numel() // inputs.shape[-1],
                self.out_features,
                self.in_features,
                self.group_size,
            )
        else:
            out = awq_inference_engine.gemm_forward_cuda_new(
                inputs, self.qweight, self.scales, self.scaled_zeros
            )  # - 8.0 * self.scales)
        out = out + self.bias if self.bias is not None else out
        return out

    def extra_repr(self) -> str:
        return (
            "in_features={}, out_features={}, bias={}, w_bit={}, group_size={}".format(
                self.in_features,
                self.out_features,
                self.bias is not None,
                self.w_bit,
                self.group_size,
            )
        )


def pseudo_quantize_tensor(
    w, n_bit=8, zero_point=True, q_group_size=-1, inplace=False, get_scale_zp=False
):
    org_w_shape = w.shape
    if q_group_size > 0:
        assert org_w_shape[-1] % q_group_size == 0
        w = w.reshape(-1, q_group_size)
    assert w.dim() == 2
    if zero_point:
        max_val = w.amax(dim=1, keepdim=True)
        min_val = w.amin(dim=1, keepdim=True)
        max_int = 2**n_bit - 1
        min_int = 0
        scales = (max_val - min_val).clamp(min=1e-5) / max_int
        zeros = (-torch.round(min_val / scales)).clamp_(min_int, max_int)

    assert torch.isnan(scales).sum() == 0
    assert torch.isnan(w).sum() == 0
    if inplace:
        (
            (w.div_(scales).round_().add_(zeros)).clamp_(min_int, max_int).sub_(zeros)
        ).mul_(scales)
    else:
        w = (
            torch.clamp(torch.round(w / scales) + zeros, min_int, max_int) - zeros
        ) * scales
    assert torch.isnan(w).sum() == 0
    w = w.reshape(org_w_shape)
    if get_scale_zp:
        return w, scales.view(w.shape[0], -1), zeros.view(w.shape[0], -1)
    else:
        return w


def is_linear(module):
    return isinstance(
        module, (torch.nn.Linear, ColumnParallelLinear, RowParallelLinear)
    )


@torch.no_grad()
def real_quantize_model_weight(model, w_bit, q_config, init_only=False):
    from .qmodule import WQLinear

    layers = model.layers
    for i in tqdm.tqdm(
        range(len(layers)),
        desc="real weight quantization..." + ("(init only)" if init_only else ""),
    ):
        layer = layers[i]
        named_linears = named_linears = {
            name: m for name, m in layer.named_modules() if is_linear(m)
        }
        # scale_activations(layer)

        for name, module in named_linears.items():
            if init_only:
                q_linear = WQLinear.from_linear(
                    module, w_bit, q_config["q_group_size"], True
                )
                q_linear.to(next(layer.parameters()).device)
                set_op_by_name(layer, name, q_linear)
            else:
                module.cuda()
                module.weight.data, scales, zeros = pseudo_quantize_tensor(
                    module.weight.data, n_bit=w_bit, get_scale_zp=True, **q_config
                )
                # scales = scales.t().contiguous()
                # zeros = zeros.t().contiguous()
                q_linear = WQLinear.from_linear(
                    module, w_bit, q_config["q_group_size"], False, scales, zeros
                )
                module.cpu()
                q_linear.to(next(layer.parameters()).device)
                set_op_by_name(layer, name, q_linear)
                torch.cuda.empty_cache()
                gc.collect()

    torch.cuda.empty_cache()
    gc.collect()
